/*
 * Project: Sudoku Explainer
 * Copyright (C) 2006-2007 Nicolas Juillerat
 * Copyright (C) 2013-2020 Keith Corlett
 * Available under the terms of the Lesser General Public License (LGPL)
 */
package diuf.sudoku;

import static diuf.sudoku.TestHelp.*;
import static diuf.sudoku.GridTestHelp.*;
import static org.junit.Assert.*;
import diuf.sudoku.Grid.ARegion;
import diuf.sudoku.Grid.Cell;
import diuf.sudoku.utils.Frmt;
import diuf.sudoku.utils.MyArrays;
import java.util.Arrays;
import org.junit.Before;
import org.junit.Test;

/**
 * Automated tests of the Grid.Cell class.
 * NOTE that Cell is an inner class of Grid, so you cannot create a Cell
 * separately, so the Cell constructors are tested in GridTests, but the
 * Cells other methods are tested here (not all one huge test-class).
 * @author Keith Corlett 2018 Apr
 */
public final class GridCellTest {

public static final String WITH_E4_8_NO_AUTO
="7.....4...2..7..83.53..8.79...5863...6..2..9...1..7..6...3529...3..49.6..49..1.35" + NL
+",189,68,1269,1369,35,,125,12,1469,,46,1469,,45,156,,,146,,,1246,16,,126,,,249,79,247,,,,,124,1247,3458,,4578,14,,34,1578,,1478,234589,89,,49,39,,258,245,,168,178,678,,,,,14,1478,1258,,2578,78,,,1278,,1278,268,,,678,6,,278,,";

//public static final String WITH_E4_8_AUTO
////                                       v
//="7.....4...2..7..83.53..8.79...5863...6.12..9...1..7..6...3529...3..49.6..49..1.35" + NL
//+",189,68,269,1369,35,,125,12,1469,,46,469,,45,156,,,146,,,246,16,,126,,,249,79,247,,,,,124,1247,3458,,4578,,,34,578,,478,234589,89,,49,39,,258,245,,168,178,678,,,,,14,1478,1258,,2578,78,,,1278,,1278,268,,,678,6,,278,,";
public static final String WITH_E4_8_AUTO
="7.....4...2..7..83.53.18.79...5863...6.12..9...1..7..6...3529...3..49.6..49.61.35" + NL
+",189,68,269,39,35,,125,12,1469,,46,469,,45,156,,,46,,,246,,,26,,,249,79,247,,,,,124,1247,3458,,4578,,,34,578,,478,234589,89,,49,39,,258,245,,168,178,678,,,,,14,1478,1258,,2578,78,,,1278,,1278,28,,,78,,,278,,";

public static final String WITH_H7_4_AUTO
="7.....4...2..7..83.53..8.79...5.63...6..2..9...1..7..6...35294..3..49.6..49..1.35" + NL
+",189,68,1269,1369,35,,125,12,1469,,46,1469,,45,156,,,146,,,1246,16,,126,,,2489,789,2478,,189,,,12,12478,3458,,4578,148,,34,1578,,1478,234589,89,,489,389,,258,25,,168,178,678,,,,,,178,1258,,2578,78,,,1278,,1278,268,,,678,68,,278,,";

private Grid grid;

@Before
public void setUp() {
	grid = new Grid();
	grid.load(LONG_STRING_3); // load doesn't "fix" missing maybes.
}

/**
 * Test Cells set(int value, boolean isAutosolving) method
 * and Grids equals(Object) method
 * and testGridConstructor_2
 */
@Test
public void testSet1() {
	// setUp() loads LONG_STRING_3
	final String testName = testName(this);
	// change one of the values on the grid: set E4=8 (correct value)
	boolean isAutosolving = false;
	grid.get("E4").set(8, 0, isAutosolving, null);
	assertGridEquals(WITH_E4_8_NO_AUTO, grid);
}

@Test
public void testSet2() {
	// setUp() loads LONG_STRING_3
	final String testName = testName(this);
	// same again with Autosolve on: set E4=8 which partially solves the puzzle
	boolean isAutosolving = true;
	Cell E4 = grid.get("E4");
	E4.set(8, 0, isAutosolving, null); // also sets 2 consequent singles
	assertGridEquals(WITH_E4_8_AUTO, grid);
}

@Test
public void testSet3() {
	// setUp() loads LONG_STRING_3
	final String testName = testName(this);
	// setting H7=4 does JUST that, there are no consequent singles
	boolean isAutosolving = true;
	grid.get("H7").set(4, 0, isAutosolving, null); // there are NO consequent singles
	assertGridEquals(WITH_H7_4_AUTO, grid);
}

/**
 * Test Cells set(int value, boolean isAutosolving) method
 * and Grids equals(Object) method
 * and testGridConstructor_2
 */
@Test
public void testRemoveFromMySiblingsMaybes() {
	// setUp() loads LONG_STRING_3
	final String testName = testName(this);
	// set E4 to 8 (its correct value)
	final int THE_VALUE = 8;
	final Cell E4 = grid.get("E4");
	E4.value = THE_VALUE;
	// remove me from my regions indexes of my maybes values
	E4.removeMyMaybesFromMyRegionsIndexes(); // uses E4.maybes // does assert/s!
	// clear my maybes
	E4.maybes.clear();
	// remove my value 8 from my siblings maybes
	E4.removeFromMySiblingsMaybes(THE_VALUE); // uses E4.value
	// check that 8 has been removed from all my siblings maybes
	for ( Cell sib : E4.siblings )
		assertEquals(false, sib.maybes.contains(THE_VALUE));
	// check that each siblings indexes have also been updated
	boolean any = false;
	for ( Cell sib : E4.siblings ) {
		if ( sib.value==0 ) {
			for ( ARegion sr : sib.regions ) {
				int si = sr.indexOf(sib);
				if ( sr.indexesOf[THE_VALUE].contains(si) ) {
					System.out.println("TRMCVFMS: Broken: Empty Sibling: "
					+sib.id+"'s index is "+si+" in "+sr
					+" whose idxsOf["+THE_VALUE+"]="+sr.indexesOf[THE_VALUE]);
					any = true;
				}
			} // next sr
		} else if ( sib.maybes.bits!=0 || sib.maybes.size!=0 ) {
			System.out.println("TRMCVFMS: WTF: Filled Sibling: "
			+sib.id+"="+sib.value
			+" maybes.bits="+sib.maybes.bits
			+" maybes.size="+sib.maybes.size);
			any = true;
		}
	} // next sibling
	assertEquals(false, any);
}

/**
 * Test the removeMyMaybesFromMyRegionsIndexes method of Grid.Cell class.
 */
@Test
public void testRemoveMyMaybesFromMyRegionsIndexes() {
	// setUp() loads LONG_STRING_3
	final String testName = testName(this);
	final boolean isNOISY = false;

	Cell E4 = grid.get("E4");

	if ( isNOISY ) {
		System.out.println("<pre>");
		System.out.println("tRMMFMRI: E4 is "+E4);
		for ( ARegion r : E4.regions )
			System.out.println("tRMMFMRI: E4 @ "+r.indexOf(E4)+" in "+r.typeName+".idxsOf[8]="+r.indexesOf[8]);
		System.out.println("</pre>");
	}
	
	// We're simulating setting E4 to 8 (it's correct value)
	E4.value = 8;
	E4.removeMyMaybesFromMyRegionsIndexes(); // {1,8,9}

	if ( isNOISY ) {
		System.out.println("<post>");
		System.out.println("tRMMFMRI: E4 is "+E4);
		for ( ARegion r : E4.regions )
			System.out.println("tRMMFMRI: E4 @ "+r.indexOf(E4)+" in "+r.typeName+".idxsOf[8]="+r.indexesOf[8]);
		System.out.println("</post>");
	}

	assertEquals(false, E4.iAmInMyRegionsIndexesOf(Values.all())); // invariant
}

/**
 * Test the removeMeFromMyRegionsIndexesOf method of Grid.Cell class.
 */
@Test
public void testRemoveMeFromMyRegionsIndexesOf1() {
	// setUp() loads LONG_STRING_3
	final String testName = testName(this);
	Cell E4 = grid.get("E4"); // E4:3{189}'s correct value is 8
	// Let's try removing 1 unused maybe
	Values maybes = new Values(9);
	E4.removeMeFromMyRegionsIndexesOf(maybes);
	assertEquals(false, E4.iAmInMyRegionsIndexesOf(maybes));
}

@Test
public void testRemoveMeFromMyRegionsIndexesOf2() {
	// setUp() loads LONG_STRING_3
	final String testName = testName(this);
	Cell E4 = grid.get("E4"); // E4:3{189}'s correct value is 8
	// Let's try removing 2 unused maybes
	Values maybes = new Values(1,9);
	E4.removeMeFromMyRegionsIndexesOf(maybes);
	assertEquals(false, E4.iAmInMyRegionsIndexesOf(maybes));

}

@Test
public void testRemoveMeFromMyRegionsIndexesOf3() {
	// setUp() loads LONG_STRING_3
	final String testName = testName(this);
	Cell E4 = grid.get("E4"); // E4:3{189}'s correct value is 8
	// Let's try setting his value and removing ALL his maybes
	Values maybes = Values.all();
	E4.value = 8; // his correct value
	E4.removeMeFromMyRegionsIndexesOf(maybes);
	assertEquals(false, E4.iAmInMyRegionsIndexesOf(maybes));
}

/**
 * Test the canSoBe method and the canNotBe method of Grid.Cell class.
 */
@Test
public void testCanSoBe() {
	// setUp() loads LONG_STRING_3
	final String testName = testName(this);
	final Values maybes = new Values(8);
	// PRE check that the grid's are equals
	assertGridEquals(LONG_STRING, grid);
	// E4:3{189}'s correct value is 8, so let's try removing 8 from his maybes,
	Cell E4 = grid.get("E4");
	E4.canNotBe(8);
	assertEquals(false, E4.iAmInMyRegionsIndexesOf(maybes));
	// and then resinstating it.
	E4.canSoBe(8);
	assertEquals(true, E4.iAmInMyRegionsIndexesOf(maybes));
	// POST check that the grid's are equals (again)
	assertGridEquals(LONG_STRING, grid);
}

/**
 * Test the isSiblingOf method of Grid.Cell class.
 */
@Test
public void testIsSiblingOf() {
	// setUp() loads LONG_STRING_3
	final String testName = testName(this);
	Cell I1 = grid.get("I1"); // eye-one
	for ( Cell cell : grid.boxs[0].cells )
		// cell is a sibling if it is on "row 1", otherwise not.
		assertEquals(cell.id, cell.y==0, !cell.notSees[I1.i]);
	Cell A9 = grid.get("A9");
	for ( Cell cell : grid.boxs[0].cells )
		// cell is a sibling if it is in "col A", otherwise not.
		assertEquals(cell.id, cell.x==0, !cell.notSees[A9.i]);
	Cell A1 = grid.get("A1");
	for ( Cell cell : grid.cells )
		if ( cell == A1 )
			assertEquals("I am not my own sibling", true, A1.notSees[cell.i]);
		else {
			// cell is a sibling if it is in "box 1", otherwise not.
			boolean expected = cell.boxId!=A1.boxId && cell.x!=A1.x && cell.y!=A1.y;
			assertEquals(cell.id, expected, A1.notSees[cell.i]);
		}
}

/**
 * Test the siblings field and therefore the getSiblingSet() method of the 
 * Grid.Cell class.
 */
@Test
public void testGetSiblingsArray() {
	// setUp() loads LONG_STRING_3
	final String testName = testName(this);
	// get the cell to test from the grid
	final Cell cell = grid.get("E4");
	// I use asserts to double-check that "the other software" I rely on to
	// build the expected sibs array contains exactly what I expect it to.
	// By the way: this is also testing the Frmt.append method.
	final ARegion box = cell.regions[Grid.BOX];
	assert "box 5".equals(box.id) && "D4 E4 F4 D5 E5 F5 D6 E6 F6".equals(Frmt.ssv(box.cells));
	final ARegion row = cell.regions[Grid.ROW];
	assert "row 4".equals(row.id) && "A4 B4 C4 D4 E4 F4 G4 H4 I4".equals(Frmt.ssv(row.cells));
	final ARegion col = cell.regions[Grid.COL];
	assert "col E".equals(col.id) && "E1 E2 E3 E4 E5 E6 E7 E8 E9".equals(Frmt.ssv(col.cells));
	// build an array of the expected siblings of cell
	// note that is all done independantly of "the software". I'm only relying
	// upon "the static stuff" to be correct, such as the cell.regions and
	// region.cells; which is all double-checked with above asserts.
	final Cell[] sibs = new Cell[20];
	int cnt = 0;
	for ( ARegion region : cell.regions ) // my box, row, col
		for ( Cell sib : region.cells )
			if ( sib!=cell && !MyArrays.contains(sibs, cnt, sib) )
				sibs[cnt++] = sib;
	// compare the expected siblings (sibs) with those in the cell itself
	// ignoring order.
	Arrays.sort(sibs, Grid.BY_ID);
	Cell[] copy = new Cell[20];
	System.arraycopy(cell.siblings, 0, copy, 0, 20);
	Arrays.sort(copy, Grid.BY_ID);
	assertArrayEquals(sibs, copy);
}

/**
 * Test the doesSiblingsContainValue method of Grid.Cell class.
 */
@Test
public void testDoesSiblingsContain() {
	// setUp() loads LONG_STRING_3
	final String testName = testName(this);
	Cell e4 = grid.get("E4");
	assertEquals(true, e4.canSeeValue(4));
	assertEquals(false, e4.canSeeValue(1));
	assertEquals(false, e4.canSeeValue(8));
	assertEquals(false, e4.canSeeValue(9));
}

}