/*
 * Project: Sudoku Explainer
 * Copyright (C) 2006-2007 Nicolas Juillerat
 * Copyright (C) 2013-2020 Keith Corlett
 * Available under the terms of the Lesser General Public License (LGPL)
 */
package diuf.sudoku;

import static diuf.sudoku.TestHelp.*;
import static org.junit.Assert.*;
import org.junit.Before;
import org.junit.Test;

import diuf.sudoku.Grid.ARegion;
import diuf.sudoku.Grid.Cell;
//import diuf.sudoku.solver.hinters.als.HdkSet;
import diuf.sudoku.utils.MyInteger;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.TreeSet;

/**
 *
 * @author Keith Corlett 2020-10-17
 */
public class IdxTest {

// self-documentary Idx constructor param
private static final boolean FULL = true;

public static final String SMELLY_CAT
="492.67.53.......7.176........1...2..9....4.1...7..5.8..3.7.8....1..9........2.4.." + NL
+",,,18,,,18,,,358,58,358,149,14,129,169,,12469,,,,34589,3458,239,89,249,2489,3568,4568,,3689,378,369,,3469,45679,,2568,358,2368,378,,3567,,567,36,246,,12369,13,,369,,469,256,,459,,145,,1569,269,12569,25678,,458,3456,,36,35678,236,25678,5678,568,589,1356,,136,,369,156789";

@Before
public void setUp() {
	grid = new Grid();
	grid.load(SMELLY_CAT);
}

private static Idx newIdx(int... indices) {
	Idx result = new Idx();
	add(result, indices);
	return result;
}

private static int add(Idx idx, int... indices) {
	int cnt = 0;
	for ( int i : indices ) {
		idx.add(i);
		++cnt;
	}
	return cnt;
}

private static int remove(Idx idx, int... indices) {
	int cnt = 0;
	for ( int i : indices ) {
		idx.remove(i);
		++cnt;
	}
	return cnt;
}

private int containsCount(Idx idx, boolean want) {
	int cnt = 0;
	for ( int i=0; i<81; ++i )
		if ( idx.contains(i) == want )
			++cnt;
	return cnt;
}

/**
 * Test of of method, of class Idx.
 */
@Test
public void testOf_Iterable() {
	// setUp() loads SMELLY_CAT
	final String testName = testName(this);
	Iterable<Cell> cells = Arrays.asList(grid.rows[0].cells);
	Idx expResult = newIdx(0,1,2,3,4,5,6,7,8);
	assertEquals(expResult, Idx.of(cells));
}

/**
 * Test of of method, of class Idx.
 */
@Test
public void testOf_GridCellArr() {
	// setUp() loads SMELLY_CAT
	final String testName = testName(this);
	Cell[] cells = grid.rows[0].cells;
	Idx expResult = newIdx(0,1,2,3,4,5,6,7,8);
	assertEquals(expResult, Idx.of(cells));
}

/**
 * Test of hashCode(a0, a1, a2), of class Idx.
 */
@Test
public void testStaticHashCode() {
	final String testName = testName(this);
	assertEquals(true, Idx.hashCode(13, 26, 42) != 0 ); // I know not and care less
}

/**
 * Test of sizeLTE method, of class Idx.
 */
@Test
public void testSizeLTE_4args() {
	final String testName = testName(this);
	assertEquals(false, Idx.sizeLTE(1,3,3, 4));
	assertEquals(true , Idx.sizeLTE(1,3,3, 5));
	assertEquals(true , Idx.sizeLTE(1,3,3, 6));
}

/**
 * Test of size method, of class Idx.
 */
@Test
public void testSize_3args() {
	final String testName = testName(this);
	assertEquals(5, Idx.size(1, 3, 3));
}

/**
 * Test of parse method, of class Idx.
 */
@Test
public void testParse() {
	final String testName = testName(this);
	Idx expResult = newIdx(0, 1, 2);
	Idx result = Idx.parse("A1, B1, C1");
	assertEquals(expResult, result);
}

///**
// * Test of toHdkSet method, of class Idx.
// */
//@Test
//public void testToHdkSet() {
//	final String testName = testName(this);
//	Idx instance = newIdx(0, 1, 2);
//	HdkSet expResult = new HdkSet();
//	expResult.add(0);
//	expResult.add(1);
//	expResult.add(2);
//	assertEquals(expResult, instance.toHdkSet());
//}

/**
 * Test of clone method, of class Idx.
 */
@Test
public void testClone() throws Exception {
	final String testName = testName(this);
	Idx instance = newIdx(42, 32, 13);
	Idx expResult = newIdx(42, 32, 13);
	assertEquals(expResult, instance.clone());
}

/**
 * Test of set method, of class Idx.
 */
@Test
public void testSet_Idx() {
	final String testName = testName(this);
	Idx src = newIdx(42, 19, 13);
	Idx instance = new Idx();
	Idx expResult = newIdx(42, 19, 13);
	assertEquals(expResult, instance.set(src));
}

/**
 * Test of set method, of class Idx.
 */
@Test
public void testSet_3args() {
	final String testName = testName(this);
	Idx instance = newIdx(42, 19, 13); // existing content is replaced
	Idx expResult = newIdx(0, 27, 28, 54, 55, 56);
	Idx result = instance.set(1, 3, 7);
//KEEP THESE: It's how I created this test-case
//System.out.println("result: "+Arrays.toString(result.toArrayA()));
	assertEquals(expResult, result);
}

/**
 * Test of set method, of class Idx.
 */
@Test
public void testSet_GridARegionArr() {
	// setUp() loads SMELLY_CAT
	final String testName = testName(this);
	ARegion[] regions = new ARegion[]{grid.rows[0]};
	Idx instance = new Idx();
	Idx expResult = newIdx(0,1,2,3,4,5,6,7,8);
	assertEquals(expResult, instance.set(regions));
}

/**
 * Test of setAnd method, of class Idx.
 */
@Test
public void testSetAnd() {
	final String testName = testName(this);
	Idx aa = newIdx(0, 2, 5, 6);
	Idx bb = newIdx(1, 3, 5, 7);
	Idx instance = new Idx();
	Idx expResult = newIdx(5);
	Idx result = instance.setAnd(aa, bb);
	assertEquals(expResult, result);
}

/**
 * Test of setAndAny method, of class Idx.
 */
@Test
public void testSetAndAny() {
	final String testName = testName(this);
	Idx aa = newIdx(0, 2, 4);
	Idx bb = newIdx(1, 3, 4);
	Idx instance = new Idx();
	Idx expResult = newIdx(4);
	assertEquals(true, instance.setAndAny(aa, bb));
	// the Idx is set as a side-effect of all its set methods... go figure.
	// The return value normally just determines if we need to read this Idx.
	assertEquals(expResult, instance);
}

/**
 * Test of setOr method, of class Idx.
 */
@Test
public void testSetOr() {
	final String testName = testName(this);
	Idx aa = newIdx(0, 2, 4);
	Idx bb = newIdx(1, 3, 5);
	Idx instance = new Idx();
	Idx expResult = newIdx(0,1,2,3,4,5);
	Idx result = instance.setOr(aa, bb);
	assertEquals(expResult, result);
}

/**
 * Test of setAndNot method, of class Idx.
 */
@Test
public void testSetAndNot() {
	final String testName = testName(this);
	Idx aa = newIdx(0, 2, 4);
	Idx bb = new Idx(FULL);
	bb.remove(4);
	Idx instance = new Idx();
	Idx expResult = newIdx(4); // it keeps whatever's missing from a full set
	Idx result = instance.setAndNot(aa, bb);
	assertEquals(expResult, result);
}

/**
 * Test of not method, of class Idx.
 */
@Test
public void testNot() {
	final String testName = testName(this);
	Idx s1 = new Idx();
	s1.not();
	for ( int i=9; i<81; ++i )
		s1.remove(i);
	Idx expResult = newIdx(0,1,2,3,4,5,6,7,8);
	assertEquals(expResult, s1);
}

/**
 * Test of clear method, of class Idx.
 */
@Test
public void testClear() {
	final String testName = testName(this);
	Idx instance = newIdx(0, 2, 4);
	Idx result = instance.clear();
	assertEquals(0, containsCount(instance, true));
	assertEquals(true, instance == result); // reference equals ok
}

/**
 * Test of fill method, of class Idx.
 */
@Test
public void testFill() {
	final String testName = testName(this);
	Idx instance = new Idx();
	Idx result = instance.fill();
	assertEquals(81, containsCount(result, true));
}

/**
 * Test of add method, of class Idx.
 */
@Test
public void testAdd_int() {
	final String testName = testName(this);
	Idx instance = new Idx();
	instance.add(0);
	instance.add(2);
	instance.add(4);
	assertEquals(3, containsCount(instance, true));
}

/**
 * Test of addAll method, of class Idx.
 */
@Test
public void testAddAll() {
	// setUp() loads SMELLY_CAT
	final String testName = testName(this);
	Cell[] cells = grid.rows[0].cells;
	Idx instance = new Idx();
	Idx expResult = newIdx(0,1,2,3,4,5,6,7,8);
	Idx result = instance.addAll(cells);
	assertEquals(expResult, result);
}

/**
 * Test of remove method, of class Idx.
 */
@Test
public void testRemove_int() {
	final String testName = testName(this);
	Idx instance = newIdx(0, 2, 4);
	instance.remove(2);
	Idx expResult = newIdx(0, 4);
	assertEquals(expResult, instance);
}

/**
 * Test of removeAll method, of class Idx.
 */
@Test
public void testRemoveAll() {
	// setUp() loads SMELLY_CAT
	final String testName = testName(this);
	List<Cell> cells = new ArrayList<>(81-9);
	for ( int i=1; i<9; ++i)
		for ( Cell cell : grid.rows[i].cells )
			cells.add(cell);
	Idx instance = new Idx(FULL);
	instance.removeAll(cells);
	int[] expResult = new int[]{0,1,2,3,4,5,6,7,8};
	assertArrayEquals(expResult, instance.toArrayA());
}

/**
 * Test of and method, of class Idx.
 */
@Test
public void testAnd() {
	final String testName = testName(this);
	Idx other = newIdx(0, 6);
	Idx instance = newIdx(0, 2, 4, 6);
	Idx expResult = newIdx(0, 6);
	Idx result = instance.and(other);
	assertEquals(expResult, result);
}

/**
 * Test of andNot method, of class Idx.
 */
@Test
public void testAndNot() {
	final String testName = testName(this);
	Idx other = new Idx(FULL);
	remove(other, 0,1,2,3,4,5,6,7,8);
	Idx instance = new Idx(FULL);
	Idx result = instance.andNot(other);
	Idx expResult = newIdx(0,1,2,3,4,5,6,7,8);
	assertEquals(expResult, result);
}

/**
 * Test of or method, of class Idx.
 */
@Test
public void testOr() {
	final String testName = testName(this);
	Idx other = newIdx(3,4,5);
	Idx instance = newIdx(0,1,2);
	Idx expResult = newIdx(0,1,2,3,4,5);
	Idx result = instance.or(other);
	assertEquals(expResult, result);
}

/**
 * Test of isEmpty method, of class Idx.
 */
@Test
public void testIsEmpty() {
	final String testName = testName(this);
	Idx instance = newIdx(1);
	assertEquals(false, instance.isEmpty());
	instance.clear();
	assertEquals(true, instance.isEmpty());
}

/**
 * Test of any method, of class Idx.
 */
@Test
public void testAny() {
	final String testName = testName(this);
	Idx instance = new Idx();
	assertEquals(false, instance.any());
	instance.add(0);
	assertEquals(true, instance.any());
	instance.remove(0);
	assertEquals(false, instance.any());
}

/**
 * Test of andAny method, of class Idx.
 */
@Test
public void testAndAny() {
	final String testName = testName(this);
	Idx other = newIdx(1);
	Idx instance = newIdx(0,1,2);
	assertEquals(true, instance.andAny(other));
	Idx expResult = newIdx(0,1,2);
	assertEquals(expResult, instance);
}

/**
 * Test of andEqualsOther method, of class Idx.
 */
@Test
public void testAndEqualsOther() {
	final String testName = testName(this);
	Idx instance = newIdx(0,1,2,3);
	Idx other = newIdx(2,3);
	assertEquals(true, instance.andEqualsOther(other));
	// it does not mutate this
	Idx expResult = newIdx(0,1,2,3);
	assertEquals(expResult, instance);
}

/**
 * Test of size method, of class Idx.
 */
@Test
public void testSize_0args() {
	final String testName = testName(this);
	Idx instance = newIdx(0,2,4);
	assertEquals(3, instance.size());
}

/**
 * Test of contains method, of class Idx.
 */
@Test
public void testContains_int() {
	final String testName = testName(this);
	Idx instance = newIdx(0, 2, 4);
	assertEquals(false, instance.contains(1));
	assertEquals(true, instance.contains(2));
}

/**
 * Test of otherThan method, of class Idx.
 */
@Test
public void testOtherThan() {
	final String testName = testName(this);
	Idx instance = newIdx(2, 4);
	assertEquals(4, instance.otherThan(2));
	Idx expResult = newIdx(2, 4);
	assertEquals(expResult, instance);
}

/**
 * Test of toArray method, of class Idx.
 */
@Test
public void testToArray() {
	final String testName = testName(this);
	int[] array = new int[3];
	Idx instance = newIdx(0, 2, 4);
	int[] expResult = new int[]{0, 2, 4};
	int[] result = instance.toArray(array);
//System.out.println(Arrays.toString(result));
	assertArrayEquals(expResult, result);
}

/**
 * Test of toArrayA method, of class Idx.
 */
@Test
public void testToArrayA() {
	final String testName = testName(this);
	Idx instance = newIdx(0,2,4,6);
	int[] expResult = new int[]{0,2,4,6};
	int[] result = instance.toArrayA();
	assertArrayEquals(expResult, result);
}

/**
 * Test of toArrayB method, of class Idx.
 * <p>
 * NB: toArrayA and toArrayB only differ in that use two separate caches of int
 * arrays, so that you can call one inside the other without fear of collision:
 * trying to the use same array to hold two different sets of values at one
 * time, which it won't, obviously.
 */
@Test
public void testToArrayB() {
	final String testName = testName(this);
	Idx instance = newIdx(0,2,4,6);
	int[] expResult = new int[]{0,2,4,6};
	int[] result = instance.toArrayB();
	assertArrayEquals(expResult, result);
}

/**
 * Test of toCellIdString method, of class Idx.
 */
@Test
public void testToCellIdString() {
	final String testName = testName(this);
	String expResult = "A1 C1 E1 G1";
	String result = Idx.toCellIdString(1+4+16+64, 0, 0);
	assertEquals(expResult, result);
}

/**
 * Test of toIndiceString method, of class Idx.
 */
@Test
public void testToIndiceString() {
	final String testName = testName(this);
	String expResult = "0 2 4 6";
	String result = Idx.toIndiceString(1+4+16+64, 0, 0);
	assertEquals(expResult, result);
}

/**
 * Test of toString method, of class Idx.
 */
@Test
public void testToString() {
	final String testName = testName(this);
	Idx instance = newIdx(0, 2, 4, 6);
	String expResult = "A1 C1 E1 G1";
	String result = instance.toString();
	assertEquals(expResult, result);
}

/**
 * Test of hashCode method, of class Idx.
 */
@Test
public void testHashCode() {
	final String testName = testName(this);
	Idx instance = newIdx(0, 4, 8, 16, 32);
	assertEquals(true, instance.hashCode() != 0); // I know not and care less
}

/**
 * Test of equals method, of class Idx.
 */
@Test
public void testEquals_3args() {
	final String testName = testName(this);
	Idx instance = newIdx(0, 2, 4);
	assertEquals(false, instance.equals(1+4+16, 1, 0));
	assertEquals(true, instance.equals(1+4+16, 0, 0));
}

/**
 * Test of equals method, of class Idx.
 */
@Test
public void testEquals() {
	final String testName = testName(this);
	Integer bogus = Integer.valueOf("0");
	Idx s1 = newIdx(0, 2, 4);
	Idx s2 = newIdx(0, 2, 4);
	assertEquals(false, s1.equals(bogus)); // no type cast exception
	assertEquals(true, s1.equals(s2));
	s1.add(1);
	assertEquals(false, s1.equals(s2));
	s1.remove(1);
	assertEquals(true, s1.equals(s2));
}

/**
 * Test of get method, of class Idx.
 */
@Test
public void testGet() {
	final String testName = testName(this);
	Idx instance = newIdx(0,2,4,6);
	assertEquals(4, instance.size()); // size and get work together in practice
	assertEquals(0, instance.get(0));
	assertEquals(2, instance.get(1));
	assertEquals(4, instance.get(2));
	assertEquals(6, instance.get(3));
}

/**
 * Test of peek method, of class Idx.
 */
@Test
public void testPeek() {
	final String testName = testName(this);
	Idx instance = newIdx(2,4,6);
	assertEquals(2, instance.peek());
	// and ensure the instance has not changed
	Idx expResult = newIdx(2,4,6);
	assertEquals(expResult, instance);
}

/**
 * Test of poll method, of class Idx.
 */
@Test
public void testPoll() {
	final String testName = testName(this);
	Idx instance = newIdx(1, 3, 5);
	// poll one off
	assertEquals(1, instance.poll());
	Idx expResult = newIdx(3, 5);
	assertEquals(expResult, instance);
	// and again
	assertEquals(3, instance.poll());
	expResult.remove(3);
	assertEquals(expResult, instance);
	// and again
	assertEquals(5, instance.poll());
	expResult.remove(5);
	assertEquals(expResult, instance);
	// and again
	assertEquals(-1, instance.poll());
	assertEquals(new Idx(), instance); // double check he's empty
}

/**
 * Test of sizeLTE method, of class Idx.
 */
@Test
public void testSizeLTE_int() {
	final String testName = testName(this);
	Idx instance = newIdx(0, 2, 4);
	assertEquals(true, instance.sizeLTE(4));
	assertEquals(true, instance.sizeLTE(3));
	assertEquals(false, instance.sizeLTE(2));
}

//toTreeSet() not used
///**
// * Test of toTreeSet method, of class Idx.
// */
//@Test
//public void testToTreeSet() {
//	// setUp() loads SMELLY_CAT
//	final String testName = testName(this);
//	Idx instance = new Idx();
//	instance.addAll(grid.rows[0].cells);
//	TreeSet<Integer> expResult = new TreeSet<>(MyInteger.list(0,1,2,3,4,5,6,7,8));
//	TreeSet<Integer> result = instance.toTreeSet();
//	assertEquals(expResult, result);
//}

}
