/*
 * Project: Sudoku Explainer
 * Copyright (C) 2006-2007 Nicolas Juillerat
 * Copyright (C) 2013-2020 Keith Corlett
 * Available under the terms of the Lesser General Public License (LGPL)
 */
package diuf.sudoku;

import org.junit.Test;
import static org.junit.Assert.*;
import static diuf.sudoku.TestHelp.*;

/**
 * Test the Indexes class, a raw data class which stores the potential
 * positions (ie the indexes) of a value in a region 0..8, and a bunch
 * of operations there-on.
 * @author Keith Corlett 2018 Mar
 */
public final class IndexesTest {
	
// ---------------- static factories ----------------

/** Test the Indexes.empty() static factory method. */
@Test
public void testEmptyStaticFactory() {
	final String testName = testName(this);
	Indexes idxs = Indexes.empty();
	assertEquals(0, idxs.size);
	assertEquals(0, idxs.bits);
}

// ---------------- constructors ----------------

/** Test of class Indexes default constructor,
 * and the basic get(int i) method. */
@Test
public void testConstructorDefault() {
	final String testName = testName(this);
	// Indexes are created full by default
	Indexes idxs = new Indexes();
	assertEquals(9, idxs.size);
	assertEquals(511, idxs.bits); // (1<<9)-1
	for ( int i=0; i<9; ++i)
		assertEquals(T, idxs.contains(i));
}

/** Test of the (int bits) constructor, of class Indexes. */
@Test
public void testConstructor_intBits() {
	final String testName = testName(this);
	Indexes idxs = new Indexes(259); //0,1,8
	check018(idxs);
}

///** Test of the (int... indexes) constructor, of class Indexes. */
//@Test
//public void testConstructor_intArr() {
//	final String testName = testName(this);
//	Indexes idxs = new Indexes(0,1,8);
//	check018(idxs);
//}

/** Test of the (String s) constructor, of class Indexes. */
@Test
public void testConstructor_Str() {
	final String testName = testName(this);
	Indexes idxs = new Indexes("018");
	check018(idxs);
}

/** Test of the copy constructor, of class Indexes. */
@Test
public void testConstructorCopy() {
	final String testName = testName(this);
	Indexes copy = new Indexes(new Indexes(259)); //0,1,8
	check018(copy);
}

// ---------------- mutators ----------------

/**
 * Test of copyFrom method, of class Indexes.
 */
@Test
public void testCopyFrom() {
	final String testName = testName(this);
	Indexes src = new Indexes(259); //0,1,8
	Indexes idxs = Indexes.empty();
	idxs.copyFrom(src);
	check018(idxs);
}

/**
 * Test of clear method, of class Indexes.
 */
@Test
public void testClear_0args() {
	final String testName = testName(this);
	Indexes idxs = new Indexes(); // filled by default
	idxs.clear();
	assertEquals(0, idxs.size);
	assertEquals(0, idxs.bits);
}

/**
 * Test of fill method, of class Indexes.
 */
@Test
public void testFill() {
	final String testName = testName(this);
	Indexes idxs = Indexes.empty();
	idxs.fill();
	assertEquals(9, idxs.size);
	assertEquals(511, idxs.bits);
}

/**
 * Test of add method, of class Indexes.
 */
@Test
public void testAdd_int() {
	final String testName = testName(this);
	Indexes idxs = new Indexes(3); // 0,1
	int i = 8;
	Indexes result = idxs.add(i); // + 8
	assert idxs == result; // add returns this, for method chaining
	check018(result);
}

/**
 * Test of clear method, of class Indexes.
 */
@Test
public void testRemove_int() {
	final String testName = testName(this);
	Indexes idxs = new Indexes(); // filled
	for ( int i=2; i<8; ++i )
		idxs.remove(i);
	check018(idxs);
}

/**
 * Test of set method, of class Indexes.
 */
@Test
public void testSet() {
	final String testName = testName(this);
	boolean isOn = false;
	Indexes idxs = new Indexes();
	for ( int i=2; i<8; ++i )
		idxs.set(i, isOn);
	check018(idxs);
}

/**
 * Test of add method, of class Indexes.
 */
@Test
public void testAdd_Indexes() {
	final String testName = testName(this);
	Indexes other = new Indexes(259); //0,1,8
	Indexes idxs = Indexes.empty();
	Indexes result = idxs.add(other);
	assert idxs == result; // add returns this, for method chaining
	check018(idxs);
}

/**
 * Test of get method, of class Indexes.
 */
@Test
public void testGet() {
	final String testName = testName(this);
	Indexes idxs = new Indexes(4); // int bits
	assertEquals(false, idxs.contains(0));
	assertEquals(false, idxs.contains(1));
	assertEquals(true, idxs.contains(2));
	for ( int i=3; i<9; ++i )
		assertEquals(false, idxs.contains(i));
	assertEquals(new Indexes("2"), idxs);
}

/**
 * Test of containsAll method, of class Indexes.
 */
@Test
public void testContainsAll() {
	final String testName = testName(this);
	Indexes idxs = new Indexes(259); //0,1,8
	assertEquals(true, idxs.containsAll(0,1,8));
	assertEquals(false, idxs.containsAll(0,1,2,8));
	assertEquals(false, idxs.containsAll(0,1,7,8));
}

/**
 * Test of isEmpty method, of class Indexes.
 */
@Test
public void testIsEmpty() {
	final String testName = testName(this);
	Indexes idxs = Indexes.empty();
	assertEquals(0, idxs.bits);
	assertEquals(0, idxs.size);
	assertEquals(true, idxs.isEmpty());
	idxs.add(0);
	assertEquals(1, idxs.bits);
	assertEquals(1, idxs.size);
	assertEquals(false, idxs.isEmpty());		
	idxs.remove(0);
	assertEquals(true, idxs.isEmpty());
	assertEquals(0, idxs.bits);
	assertEquals(0, idxs.size);
}

/**
 * Test of size method, of class Indexes.
 */
@Test
public void testSize() {
	final String testName = testName(this);
	Indexes idxs = new Indexes(259); //0,1,8
	assertEquals(3, idxs.size); // chech that the constructor correctly sets the size
	idxs.size = 0; // setting another objects field is naughty (except in unit tests (and readers)).
	assertEquals(3, idxs.size()); // sets the size field as a side-effect
	assertEquals(3, idxs.size); // check that the field has been set.
	// note that size() method is a hang-over, we used to have a justOr
	// method that, umm, just or-ed the bits with some new new ones in a
	// loop, leaving the Integer.bitCount (expensive) until after the loop.
	// I'm keeping it because it's useful for checking the invariant:
	// {@code idxs.size == idxs.size();} even though transivity is broken,
	// ie {@code size() == size;} has become a tautology because size() is
	// evalutated first, which sets and returns the size field.
}

/**
 * Test of toBinaryString method, of class Indexes.
 */
@Test
public void testToBinaryString() {
	final String testName = testName(this);
	assertEquals("000000011", new Indexes(3).toBinaryString()); //0,1 // note the leading zeros
	assertEquals("100000011", new Indexes(259).toBinaryString()); //0,1,8
}

/**
 * Test of first method, of class Indexes.
 */
@Test
public void testFirst() {
	final String testName = testName(this);
	assertEquals(-1, Indexes.empty().first());
	assertEquals(0, new Indexes(1).first()); //0
	assertEquals(0, new Indexes(259).first()); //0,1,8
	assertEquals(8, new Indexes(256).first()); //8
}

/**
 * Test of last method, of class Indexes.
 */
@Test
public void testLast() {
	final String testName = testName(this);
	assertEquals(-1, Indexes.empty().last());
	assertEquals(0, new Indexes(1).last()); //0
	assertEquals(8, new Indexes(259).last()); //0,1,8
}

/**
 * Test of next method, of class Indexes.
 */
@Test
public void testNext() {
	final String testName = testName(this);
	Indexes idxs = new Indexes(259); //0,1,8
	assertEquals(0, idxs.next(0));
}

/**
 * Test of toArray method, of class Indexes.
 */
@Test
public void testToArray_intArr() {
	final String testName = testName(this);
	int[] array = new int[9];  int n; // variables to be resused within a loop (in the real code)
	Indexes idxs = new Indexes(259); //0,1,8
	n = idxs.toArray(array);
	assertEquals(3, n);
	int[] expResult = {0,1,8,0,0,0,0,0,0};
	assertArrayEquals(expResult, array);
	// clear the 1 and retest
	idxs.remove(1);
	expResult[1] = 8;
	expResult[2] = 0;
	n = idxs.toArray(array);
	assertEquals(2, n);
	assertArrayEquals(expResult, array);
}

/**
 * Test of toArray method, of class Indexes.
 */
@Test
public void testToArray_0args() {
	final String testName = testName(this);
	Indexes idxs = new Indexes(259); //0,1,8
	int[] array = idxs.toArray();
	assertEquals(3, array.length);
	assertArrayEquals(new int[]{0,1,8}, array);
	// clear the 1 and retest
	idxs.remove(1);
	array = idxs.toArray();
	assertEquals(2, array.length);
	assertArrayEquals(new int[]{0,8}, array);
}

//if you want to test it go make the method public, test it, and put it back
//to private; I refuse to expost methods just to make them testable. Nuts!
///**
// * Test Indexes.toValuesArray static method.
// */
//@Test
//public void testToValuesArray() {
//	final String testName = testName(this);
//	assertArrayEquals(new int[]{0,1,8}, Indexes.toValuesArray(259));
//}

/**
 * Test Indexes.toShiftedArray static method.
 */
@Test
public void testToShiftedArray() {
	final String testName = testName(this);
	assertArrayEquals(new int[]{1,2,256}, Indexes.toShiftedArray(259));
}

/**
 * Test of appendTo method, of class Indexes.
 */
@Test
public void testAppendTo_StringBuilder() {
	final String testName = testName(this);
	Indexes idxs = new Indexes(259); //0,1,8
	StringBuilder sb = new StringBuilder();
	// note that appendTo (oddly) returns the sb that was passed to it.
	assertEquals("018", idxs.appendTo(sb).toString());
}

/**
 * Test of appendTo method, of class Indexes.
 */
@Test
public void testAppendTo_3args() {
	final String testName = testName(this);
	Indexes idxs = new Indexes(259); //0,1,8
	StringBuilder sb = new StringBuilder();
	String sep = ", ";
	String lastSep = " and ";
	assertEquals("0, 1 and 8", idxs.appendTo(sb, sep, lastSep).toString());
	// KRC 2018-05-02 fixed bug --> "0 and 1 and 8"
	// Yeah, I've finally used a crap-test-case to fix a bug in my code, and
	// not vice-versa. I've only written 42,000-odd of the little ____ers, so
	// I suppose it was bound to happen eventually.
}

/**
 * Test of toString method, of class Indexes.
 */
@Test
public void testToString() {
	final String testName = testName(this);
	assertEquals("018", new Indexes(259).toString()); //0,1,8
}

/**
 * Test of equals method, of class Indexes.
 */
@Test
public void testEquals_Object() {
	final String testName = testName(this);
	Indexes expIdxs = new Indexes(0, 1, 8);
	assertNotEquals(expIdxs, null);
	Object that = new Indexes(259);
	assertEquals(expIdxs, that);
}

/**
 * Test of equals method, of class Indexes.
 */
@Test
public void testEquals_Indexes() {
	final String testName = testName(this);
	Indexes that = new Indexes(259);
	assertEquals(new Indexes(0, 1, 8), that);
}

/**
 * Test of hashCode method, of class Indexes.
 */
@Test
public void testHashCode() {
	final String testName = testName(this);
	assertEquals(259, new Indexes(259).hashCode()); //0,1,8
}

private void check018(Indexes idxs) {
	assertEquals(3, idxs.size);
	assertEquals(259, idxs.bits); // (1<<0) | (1<<1) | (1<<8)
	assertEquals(T, idxs.contains(0));
	assertEquals(T, idxs.contains(1));
	for ( int i=2; i<8; ++i)
		assertEquals(F, idxs.contains(i));
	assertEquals(T, idxs.contains(8));
	assertEquals(new Indexes(259), idxs);
}
	
}
