/*
 * Project: Sudoku Explainer
 * Copyright (C) 2006-2007 Nicolas Juillerat
 * Copyright (C) 2013-2020 Keith Corlett
 * Available under the terms of the Lesser General Public License (LGPL)
 */
package diuf.sudoku;

import diuf.sudoku.Grid.Cell;
import org.junit.Test;
import static org.junit.Assert.*;
import static diuf.sudoku.TestHelp.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Tests the Pots class, which extends a Map of Cell => Values to represent 
 * the values of a set of cells.
 * @author Keith Corlett
 */
public final class PotsTest {

private static final String PUZZLE
="....41...9..5...81.7...96.....4.....3..6..14..8.1....3......83579.........2....1.";
//1465#C:\Users\User\Documents\SodukuPuzzles\MagicTour\top1465.d3.mt

// ---------------- constructors ----------------

/**
 * Test of Pots Default Constructor.
 */
@Test
public void testConstructor_Default() {
	final String testName = testName(this);
	Pots pots = new Pots();
	assertEquals(true, pots.isEmpty());
	assertEquals(8, pots.getCapacity());
}

/**
 * Test of Pots(Cell cell, Values values) constructor,
 * and the firstKey() and firstValue() methods.
 */
@Test
public void testConstructor_Cell_Values() {
	final String testName = testName(this);
	grid = new Grid(PUZZLE);
	Cell cell = grid.get("A1");
	Values values = new Values(6);
	Pots pots = new Pots(cell, values);
	assertEquals(1, pots.size());
	assertEquals(cell, pots.firstKey());
	assertSame(values, pots.firstKeyValue());
}

/**
 * Test of Pots(Cell cell, Values values) constructor.
 */
@Test
public void testConstructor_Cell_int() {
	final String testName = testName(this);
	grid = new Grid(PUZZLE);
	Cell cell = grid.get("A1");
	Values values = new Values(6);
	Pots pots = new Pots(cell, 6);
	assertEquals(1, pots.size());
	assertEquals(cell, pots.firstKey());
	assertEquals(values, pots.firstKeyValue());
}

/**
 * Test of Pots Copy Constructor.
 */
@Test
public void testConstructor_Copy() {
	final String testName = testName(this);
	grid = new Grid(PUZZLE);
	Cell cell = grid.get("A1");
	Values values = new Values(6);
	Pots pots = new Pots(new Pots(cell, values)); // 2 constructors
	assertEquals(1, pots.size());
	assertEquals(cell, pots.firstKey());
	assertSame(values, pots.firstKeyValue());
}

/**
 * Test of Pots(int initialCapacity, float loadFactor) Constructor.
 */
@Test
public void testConstructor_capacity_load() {
	final String testName = testName(this);
	Values values = new Values(6);
	Pots pots = new Pots(32, 1F);
	assertEquals(0, pots.size());
	assertEquals(32, pots.getCapacity());
}

/**
 * Test of Pots(Cell cell, int value) Constructor.
 */
@Test
public void testConstructor_CellArr_Values() {
	final String testName = testName(this);
	grid = new Grid(PUZZLE);
	Cell[] cells = grid.get("A1","B1");
	Values values = new Values(6,8);
	Pots pots = new Pots(cells, values);
	assertArrayEquals(cells, pots.keySet().toArray());
	assertEquals(values, pots.get(grid.get("A1")));
}

// ---------------- constructors ----------------

/**
 * Test of putAll method, of class Pots.
 */
@Test
public void testPutAll2() {
	final String testName = testName(this);
	grid = new Grid(PUZZLE);
	Pots pots = new Pots(8, 1F);
	Cell a1=grid.get("A1");
	Values av=new Values(6,8);
	pots.putAll2(new Pots(a1, av));
	assertEquals(av, pots.get(a1));
}

/**
 * Test of addsert method, of class Pots.
 */
@Test
public void testUpsert_Cell_Values() {
	final String testName = testName(this);
	grid = new Grid(PUZZLE);
	Cell a1=grid.get("A1");
	Pots pots = new Pots(a1, new Values(6));
	pots.upsert(a1, newValues(2,5,8));
	assertEquals(newValues(2,5,6,8), pots.get(a1));
}

/**
 * Test of addsert method, of class Pots.
 */
@Test
public void testUpsert_Cell_int() {
	final String testName = testName(this);
	grid = new Grid(PUZZLE);
	Cell a1=grid.get("A1");
	Pots pots = new Pots(a1, newValues(2,5,8));
	pots.upsert(a1, 6);
	assertEquals(newValues(2,5,6,8), pots.get(a1));
}

/**
 * Test of addsertAll method, of class Pots.
 */
@Test
public void testUpsertAll() {
	final String testName = testName(this);
	grid = new Grid(PUZZLE);
	Pots pots = new Pots(grid.get("A1"), new Values(2,8));
	Pots toUpsert = new Pots();
	toUpsert.put(grid.get("A1"), new Values(5,6)); // combine values
	toUpsert.put(grid.get("B1"), newValues(2,3,5,6)); // put
	pots.upsertAll(toUpsert);
	assertEquals(newValues(2,5,6,8), pots.get(grid.get("A1")));
	assertEquals(newValues(2,3,5,6), pots.get(grid.get("B1")));
}

/**
 * Test of populate method, of class Pots.
 */
@Test
public void testPopulate_Cells_idxs_value() {
	final String testName = testName(this);
	grid = new Grid(PUZZLE);
	Pots pots = new Pots();
	pots.populate(grid.get("A1","B1","C1"), 3, 9);
	assertEquals(new Values(9), pots.get(grid.get("A1")));
	assertEquals(new Values(9), pots.get(grid.get("B1")));
	assertEquals(null, pots.get(grid.get("C1")));
}

/**
 * Test of populate method, of class Pots.
 */
@Test
public void testPopulate_Cells_idxs_shiftedValue_value() {
	final String testName = testName(this);
	final int value = 6;
	final int shiftedValue = Values.VSHFT[value];
	Pots pots = new Pots(); // 7 = the first 3
	pots.populate(grid.rows[0].cells, 7, shiftedValue, value);
	assertArrayEquals(grid.get("A1","B1","C1"), pots.keySet().toArray());
	assertArrayEquals(new int[]{value}, pots.get(grid.get("A1")).toArrayNew());
}

/**
 * Test of removeAll method, of class Pots.
 */
@Test
public void testRemoveAll_Iterable() {
	final String testName = testName(this);
	Pots pots = new Pots();
	pots.put(grid.get("A1"), newValues(2,5,6,8));
	pots.put(grid.get("B1"), newValues(2,3,5,6));
	List<Cell> cellsToRemove = new ArrayList<>(Arrays.asList(grid.get("A1"),grid.get("C1")));
	pots.removeAll(cellsToRemove);
	assertEquals(1, pots.size());
	assertEquals(newValues(2,3,5,6), pots.get(grid.get("B1")));
}

/**
 * Test of removeAll method, of class Pots.
 */
@Test
public void testRemoveAll_Pots() {
	final String testName = testName(this);
	Pots pots = new Pots();
	pots.put(grid.get("A1"), newValues(2,5,6,8));
	pots.put(grid.get("B1"), newValues(2,3,5,6));
	Pots potsToRemove = new Pots();
	potsToRemove.put(grid.get("A1"), new Values(5,6));
	pots.removeAll(potsToRemove);
	assertEquals(2, pots.size());
	assertEquals(newValues(2,8), pots.get(grid.get("A1")));
	assertEquals(newValues(2,3,5,6), pots.get(grid.get("B1")));
}

/**
 * Test of totalSize method, of class Pots.
 */
@Test
public void testCount() {
	final String testName = testName(this);
	Pots pots = new Pots();
	pots.put(grid.get("A1"), newValues(2,5,6,8));
	pots.put(grid.get("B1"), newValues(2,3,5,6));
	assertEquals(8, pots.totalSize());
}

/**
 * Test of toString method, of class Pots.
 */
@Test
public void testToString() {
	final String testName = testName(this);
	Pots pots = new Pots();
	pots.put(grid.get("A1"), newValues(2,5,6,8));
	pots.put(grid.get("B1"), newValues(2,3,5,6));
	assertEquals("A1-2568, B1-2356", pots.toString());
}

}
