/*
 * Project: Sudoku Explainer
 * Copyright (C) 2006-2007 Nicolas Juillerat
 * Copyright (C) 2013-2020 Keith Corlett
 * Available under the terms of the Lesser General Public License (LGPL)
 */
package diuf.sudoku;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;
import diuf.sudoku.Grid.ARegion;
import diuf.sudoku.Grid.Cell;
import diuf.sudoku.solver.AHint;
import diuf.sudoku.solver.IPreparer;
import diuf.sudoku.solver.LogicalSolver;
import diuf.sudoku.solver.LogicalSolverFactory;
import diuf.sudoku.solver.accu.HintsAccumulator;
import diuf.sudoku.solver.accu.IAccumulator;
import diuf.sudoku.solver.hinters.AHinter;
//import diuf.sudoku.solver.hinters.als.HdkSet;
import diuf.sudoku.utils.Html;
import diuf.sudoku.utils.MyLinkedHashSet;
import diuf.sudoku.utils.ResourceException;
import java.io.File;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;


/**
 * Constants to be static imported into other Test classes
 * @author Keith Corlett
 */
public final class TestHelp {

// for the creation of Chainers
public static final boolean F=false, T=true;

// A shortcut to produce/reproduce all the checkHtml sample HTML files.
public static final boolean REPRODUCE_ALL_TESTS = false; // @check false

public static final String NL = diuf.sudoku.utils.Frmt.NL;

public static final String DOCS = "C:\\Users\\User\\Documents\\";
public static final String PROJECTS = DOCS+"NetBeansProjects\\";
public static final String PROJECT = PROJECTS+"DiufSudoku\\";
public static final String TEST_DATA_DIR = PROJECT+"test\\data\\";

public static final File PUZZLES_DIR = new File(DOCS+"SodukuPuzzles");

public static boolean REBUILD_GRID = false;

public static String testName(Object caller) {
	StackTraceElement ste = Thread.currentThread().getStackTrace()[2];
	String testName = caller.getClass().getSimpleName()+"."+ste.getMethodName();
	System.out.println();
	System.out.println(testName);
	return testName;
}

// the global hints list
public final static LinkedList<AHint> HINTS = new LinkedList<>();

// the global hints accumulator (hinters use this to add to the hints list)
public static IAccumulator accu = new HintsAccumulator(HINTS);

// the global grid
public static Grid grid;

// =============================== TOWEL LOCATED ==============================

public static Values newValues(int... vals) {
	return new Values(vals); // trixie: varargs ONLY in test-cases!
}

// ================================ GET KNOTTED ===============================

public static int getHints(String puzzle, AHinter hinter) {
	final boolean multiline = puzzle.indexOf('\r')>-1 || puzzle.indexOf('\n')>-1;
	HINTS.clear(); // for when get hints is called repeatedly
	grid = new Grid();
	grid.load(puzzle);
	if ( REBUILD_GRID ) {
		grid.rebuildAllRegionsEmptyCellCounts();
		grid.rebuildAllRegionsContainsValues();
		grid.rebuildAllRegionsIdxsOfAllValues();
	}
	if ( multiline && !grid.isMaybesLoaded )
		fail("FAILED TO LOAD MAYBES: "+puzzle);
	// PRE_CHECK that the loaded puzzle is the same as the given puzzle.
	String loaded = multiline ? grid.toString() : grid.toShortString();
	if ( !puzzle.equals(loaded) )
		System.out.println("PRE_CHECK EXPECTED:\n"+puzzle+"\nACTUAL:\n"+loaded);
	assertEquals(puzzle, loaded);
	// get the freakin hints already already.
	// nb: accu is a HintsAccumulator populating HINTS
	if ( !hinter.findHints(grid, accu) ) {
		if ( HINTS.size() > 0 ) {
			for ( AHint hint : HINTS )
				System.out.println(hint);
			fail(hinter+" returned false but HINTS.size()=="+HINTS.size());
		} else
			System.out.println("WARNING: "+hinter+" found nothing.");
	}
	return HINTS.size();
}

// Prepare the hinter and then run getHints.
// No need for the pre-check in getHints(puzzle, hinter).
// nb: this relies on all IPreparer's being AHinters
public static void getHints(String puzzle, IPreparer hinter, boolean prepare) {
	HINTS.clear(); // for when get hints is called repeatedly
	grid = new Grid();
	grid.load(puzzle);
	// get the LogicalSolver.
	LogicalSolver logicalSolver = LogicalSolverFactory.get();
	// set grid.solution and prepares all the hinters in the solver, but not my
	// seperate hand-rolled hinter.
	logicalSolver.prepare(grid);
	// so I prepare my hinter seperately, using the existing grid.solution
	// NB: logicalSolver normally "contains" all the hinters, it's only the
	// test-cases who use "foreign" hinters having prep-troubles like this.
	// Alternately, LogicalSolver could expose getWantedHinter(Tech) to the
	// test-cases, but that sounds like work to me.
	hinter.prepare(grid, logicalSolver);
	// get the hints
	if ( !((AHinter)hinter).findHints(grid, accu) )
		fail(""+hinter+" found nothing.");
}

// ================================ FIND S__T =================================

/**
 * Finds the first AHint whose toString starts with 'what'.
 * @param hints to search in.
 * @param target to look for,
 * @return the first AHint whose toString startsWith 'what'.
 */
public static AHint find(List<AHint> hints, String target) {
	for ( AHint hint : hints )
		if ( hint.toFullString().startsWith(target) )
			return hint;
	// YOU ONLY GET HERE WHEN NO HINT IS FOUND!
	System.out.format("TARGET HINT NOT FOUND: %s\n", target);
	// dump the hints that were found
	for ( AHint hint : hints ) // print hashCode coz clashes may be an issue
		System.out.format("%12d\t%s\n", hint.hashCode(), hint.toFullString());
	fail("No hint starting with: \""+target+"\"");
	return null;
}

public static AHint find(List<AHint> hints, String target, String className) {
	for ( AHint hint : hints )
		if ( hint.getClass().getSimpleName().equals(className)
		  && hint.toString().startsWith(target) )
			return hint;
	// YOU ONLY GET HERE WHEN NO HINT IS FOUND!
	// dump the hints that were found
	for ( AHint hint : hints ) // print with the class's simple name
		System.out.format("%-30s\t%s\n", hint.getClass().getSimpleName(), hint.toString());
	fail("No className=\""+className+"\" hint starting with: \""+target+"\"");
	return null;
}

// ================================ CHECK S__T ================================

// check selected cells: idsString is "A2, H3, G3, I3"
public static void checkAquaCells(AHint hint, String idsString) {
	if ( idsString.indexOf(", ") > -1 )
		checkAquaCells(hint, false, idsString.split(", "));
	else // presumably it's just the one cell id
		checkAquaCells(hint, false, idsString);
}
// check selected cells: ids is an args-array: "A2", "H3",
// dummy just differentiates the method signature from overloaded counterparts.
public static void checkAquaCells(AHint hint, boolean dummy, String... ids) {
	Cell[] cells = grid.get(ids);
	Set<Cell> expAquaCells = new MyLinkedHashSet<>(cells);
	assertEquals(expAquaCells, hint.getAquaCells(0));
}

// check the clue text
public static void checkClue(AHint hint, String expClueHtml) {
	assertEquals(expClueHtml, hint.getClueHtml(true));
}

// ======================== ASSERT YOUR ASS RIGHT OFF! ========================

public static boolean assertHintEqualsStoredHintsHtml(
		Object caller, AHint hint, String filename, boolean produce) {
	return assertHintEqualsStoredHintsHtml(caller, hint, filename, produce, true);
}
public static boolean assertHintEqualsStoredHintsHtml(Object caller, AHint hint
		, String filename, boolean produce, boolean failIfProduce) {
	boolean result = false; // ie test really should have failed
	String actualHtml = hint.toHtml();
	if ( produce ) {
		try {
			// nb: false makes Html.save not cache this file
			Html.save(caller, filename, actualHtml, false);
			System.out.println("PRODUCED "+filename+" <== "+hint);
		} catch (ResourceException ex) {
			fail(ex.getClass().getSimpleName()+": "+ex.getMessage());
		}
		if ( failIfProduce )
			fail("failed because PRODUCE is true!");
	} else {
		String expHtml = null;
		try {
			expHtml = Html.actuallyLoad(caller, filename, null); // NO cache
		} catch (ResourceException ex) {
			fail(ex.getClass().getSimpleName()+": "+ex.getMessage());
		}
		//System.out.println(htmlName+" <== "+hint);
		assert expHtml != null;
		if ( !expHtml.equals(actualHtml) ) {
			System.out.format("EXPECTED: %s\n%s\n", filename, expHtml);
			System.out.format("ACTUAL: %s\n%s\n", hint.toString(), actualHtml);
		}
		assertEquals(expHtml, actualHtml);
		result = true;
	}
	return result;
}

public static void assertHintEqualsStoredHintsHtml(Object caller, String puzzle
		, AHinter hinter, String htmlName, String hintToString
		, boolean produce) {
	getHints(puzzle, hinter);
	AHint hint = find(HINTS, hintToString);
	TestHelp.assertHintEqualsStoredHintsHtml(caller, hint, htmlName, produce);
}

public static void assertGridEquals(Grid expected, Grid actual) {
	if ( !expected.equals(actual) ) {
		System.out.println("expected  : \n"+expected+"\n"
						 + "actual    : \n"+actual+"\n");
		fail("unexpected result");
	}
}
public static void assertGridEquals(String expected, Grid actual) {
//	Grid expectedGrid = new Grid();
//	expectedGrid.load(expected);
//	assertGridEquals(expectedGrid, actual);
	String actualString = actual.toString();
	if ( !expected.equals(actualString) ) {
		System.out.println("expected  : \n"+expected+"\n"
						 + "actual    : \n"+actualString+"\n");
		fail("unexpected result");
	}
}

// =============================== CELL HELPERS ===============================

/** Get a String[] of cell-ids from the given csv. */
public static String[] ids(String csv) {
	List<String> list = new LinkedList<>();
	for ( String id : csv.split(" *, *") )
		list.add(id);
	return list.toArray(new String[list.size()]);
}

public static Set<Cell> cellSet(String idsCSV) {
	if ( idsCSV==null || idsCSV.length()==0 )
		return null;
	return grid.getCSVSet(idsCSV);
}

// ============================== REGION HELPERS ==============================

/** Return an array of regions for a CSV list of region IDs. */
public static ARegion[] regions(String idsCsv) {
	List<ARegion> list = new LinkedList<>();
	for ( String id : idsCsv.split(" *, *") )
		list.add(grid.region(id));
	return list.toArray(new ARegion[list.size()]);
}

/**
 * Parse an interleaved-regions array (a Fish regions array) from the string
 * produced by {@link diuf.sudoku.utils.Frmt#basesAndCovers} into an
 * interleaved-regions array (with evens as bases, and covers as odds).
 * <p>
 * I'm tempted to move this method to Frmt, but it's currently only used in
 * test-cases, so it's here in TestHelp, which is static imported; for now.
 *
 * @param basesAndCovers for example:<pre>
 * "row 2, row 5, row 8, row 9 and col A, col C, col F, col H"
 *  bases .................... and covers ...................</pre>
 * @return a {@code new ARegion[]} with evens as bases, and covers as odds
 */
public static ARegion[] parseBasesAndCovers(String basesAndCovers) {
	if ( basesAndCovers==null || basesAndCovers.length()<5 )
		return new ARegion[0]; // an empty array
	// remove " (fins C3, C4)" from the tail of Sashimi fish bases
	basesAndCovers = basesAndCovers.replaceAll(" \\(.*\\)", "");
	String[] sets = basesAndCovers.split(" and ");
	String[] bases = sets[0].split(", ");
	String[] covers = sets[1].split(", ");
	assert bases.length == covers.length;
	ARegion[] array = new ARegion[bases.length*2];
	int index = 0;
	for ( String id : bases ) {
		array[index] = grid.region(id);
		index += 2;
	}
	index = 1;
	for ( String id : covers ) {
		array[index] = grid.region(id);
		index += 2;
	}
	return array;
}

// =============================== POTS HELPERS ===============================

/**
 * Build new Pots (Potential Values of Cells) used only by test cases.
 * @param v the value.
 * @param ids an arguments list of A1..I9 id's.
 * @return a new Pots
 */
public static Pots newPots(int v, String... ids) {
	int n = ids.length;
	Pots pots = new Pots(Math.max(Pots.MIN_CAPACITY,n), 1F);
	for ( int i=0; i<n; ++i )
		pots.put(grid.get(ids[i]), new Values(v));
	return pots;
}

/**
 * Build new Pots (Potential Values of Cells) used only by test cases.
 * @param v the value.
 * @param cells an arguments list of Cell.
 * @return a new Pots
 */
public static Pots newPots(int v, Cell... cells) {
	int n = cells.length;
	Pots pots = new Pots(Math.max(Pots.MIN_CAPACITY,n), 1F);
	for ( int i=0; i<n; ++i )
		pots.put(cells[i], new Values(v));
	return pots;
}

///**
// * Build new Pots (Potential Values of Cells) used only by test cases.
// * @param values Values of each cell.
// * @param ids an arguments list of A1..I9 id's.
// * @return a new Pots
// */
//public static Pots newPots(Values values, String... ids) {
//	int n = ids.length;
//	Pots pots = new Pots(Math.max(Pots.MIN_CAPACITY,n), 1F);
//	for ( int i=0; i<n; ++i )
//		pots.put(grid.get(ids[i]), new Values(values));
//	return pots;
//}

/**
 * Build new Pots (Potential Values of Cells) used only by test cases.
 * @param s Example 1: "A2-19 B3-15 A4-89 B6-589",<br>
 * Example 2: "C1-2, D3-2, D9-2, F1-2" (from toString)
 * @return a new Pots
 */
public static Pots newPots(String s) {
	if ( s==null || s.length() == 0 )
		return new Pots();
	// replace each comma with a space;
	// replace two-or-more-spaces with a space;
	// split on a space
	String[] fields = s.replaceAll(","," ").replaceAll("  +", " ").split(" ");
	int n = fields.length;
	Pots pots = new Pots(Math.max(Pots.MIN_CAPACITY,n), 1F);
	for ( int i=0; i<n; ++i ) {
		String field = fields[i];
		String cellId = field.substring(0, 2);
		Cell cell = grid.get(cellId);
		Values values = field.length() > 2
				? new Values(field.substring(3))
				: new Values();
		pots.put(cell, values);
	}
	return pots;
}

// =============================== LINKS HELPERS ==============================

public static List<Link> link(String... ids) {
	List<Link> links = new ArrayList<>(ids.length);
	int n = ids.length;
	for ( int i=0; i<n; ++i )
		links.add(new Link(grid.get(ids[i]), 0, grid.get(ids[(i+1)%n]), 0));
	return links;
}

// ================================ HdkSet CRAP ===============================

//public static HdkSet hdkSetAllBits(Grid grid, int bits) {
//	HdkSet result = new HdkSet();
//	for ( Cell cell : grid.cells )
//		if ( (cell.maybes.bits & bits) == bits )
//			result.add(cell.i);
//	return result;
//}
//
//public static HdkSet hdkSetAnyBits(Grid grid, int bits) {
//	HdkSet result = new HdkSet();
//	for ( Cell cell : grid.cells )
//		if ( (cell.maybes.bits & bits) != 0 )
//			result.add(cell.i);
//	return result;
//}

// ---- Idx equivalents of the HdkSet CRAP ----

public static Idx idxAllBits(Grid grid, int bits) {
	Idx result = new Idx();
	for ( Cell cell : grid.cells )
		if ( (cell.maybes.bits & bits) == bits )
			result.add(cell.i);
	return result;
}

public static Idx idxAnyBits(Grid grid, int bits) {
	Idx result = new Idx();
	for ( Cell cell : grid.cells )
		if ( (cell.maybes.bits & bits) != 0 )
			result.add(cell.i);
	return result;
}

// ---- pre-existing Idx CRAP ----

public static Idx idxFromCsv(String csv) {
	Idx result = new Idx();
	for ( String id : csv.split("[ ,]+") )
		result.add(Grid.indice(id));
	return result;
}

// ================================ CONSTRUCTOR ===============================

private TestHelp() {} // Never used

}
