/*
 * Project: Sudoku Explainer
 * Copyright (C) 2006-2007 Nicolas Juillerat
 * Copyright (C) 2013-2020 Keith Corlett
 * Available under the terms of the Lesser General Public License (LGPL)
 */
package diuf.sudoku;

import org.junit.Test;
import static org.junit.Assert.*;
import static diuf.sudoku.TestHelp.*;
import java.util.LinkedList;

/**
 * Test the Values class, which is raw data class that holds a set of the
 * values 1..9.
 * @author Keith Corlett 2018 Mar
 */
public final class ValuesTest {

// ---------------- static factory methods ----------------

/**
 * Test of all method, of class Values
 * and also the toString() method.
 */
@Test
public void testStatic_All() {
	final String testName = testName(this);
	assertEquals("123456789", Values.all().toString());
}

/**
 * Test of none method, of class Values
 * and also the toString() method.
 */
@Test
public void testStatic_None() {
	final String testName = testName(this);
	assertEquals("-", Values.none().toString());
}

// ---------------- static helper methods ----------------

/**
 * Test of common method, of class Values.
 */
@Test
public void testStatic_Common() {
	final String testName = testName(this);
	// normal usage: first a valid-case test
	Values[] valueses = {new Values(1,2), new Values(2,3), new Values(1,3)};
	assertEquals(newValues(1,2,3), Values.common(valueses, 3));
	// normal usage: this returns null coz there are 4 distinct values:
	valueses = new Values[]{new Values(1,9), new Values(2,3), new Values(1,3)};
	assertEquals(null, Values.common(valueses, 3));

	// bad usage: this returns null coz the first Values doesn't have atleast 2 values.
	valueses = new Values[]{new Values(1), new Values(2,3), new Values(1,3)};
	assertEquals(null, Values.common(valueses, 3));
	// bad usage: this returns null coz the first Values has more than 3 values.
	valueses = new Values[]{newValues(1,2,3,4), new Values(2,3), new Values(1,3)};
	assertEquals(null, Values.common(valueses, 3));
}

// ---------------- constructors ----------------

/** 
 * Test of class Values default constructor.
 * Values are created empty by default.
 */
@Test
public void testConstructor_Default() {
	final String testName = testName(this);
	Values vals = new Values();
	assertEquals(0, vals.size);
	assertEquals(0, vals.bits); // 1<<(v-1)
}

/** Test of class Values(int value) constructor. */
@Test
public void testConstructor_intValue() {
	final String testName = testName(this);
	// first a valid constructor
	Values vals = new Values(1);
	assertEquals(1, vals.size);
	assertEquals(1, vals.bits); // 1<<(v-1)
	// 0 is not a valid value, so we just have to throw something, which
	// may as well be an assert, because that's erased in production.
	// We're assuming that the tests have a 100% code coverage, which they
	// don't, but hey, it's still better than beating pleasemen to death
	// for a living.
	try {
		vals = new Values(0);
		fail("No AssertionError");
		System.out.format("You can't get here: %s\n", vals);
	} catch(AssertionError ex) {
		// it passed
	}
}

/** Test of class Values(int bits, boolean dummy) constructor. */
@Test
public void testConstructor_intBits_booleanDummy() {
	final String testName = testName(this);
	// first a valid constructor
	Values vals = new Values(259, false); //1,2,9
	assertEquals(3, vals.size);
	assertEquals(259, vals.bits); //1,2,9
	// now an invalid one
	try {
		vals = new Values((1<<9), false); // 10 bits exceeds max of 9 bits
		fail("No IllegalArgumentException");
		System.out.format("You can't get here: %s\n", vals);
	} catch(IllegalArgumentException ex) {
		// it passed: new Values should throw an IllegalArgumentException: bits 512 not 0..511
	}
	// another invalid one
	try {
		vals = new Values(-1, false); // -1 bits preceeds min of 0
		fail("No IllegalArgumentException");
		System.out.format("You can't get here: %s\n", vals);
	} catch(IllegalArgumentException ex) {
		// it passed: new Values should throw an IllegalArgumentException: bits -1 not 0..511
	}
}

/** Test of class {@code Values(int... values)} constructor. */
@Test
public void testConstructor_intArr() {
	Values vals = newValues(1,2,9); //bits=259
	assertEquals(3, vals.size); 
	assertEquals(259, vals.bits); 
	assertEquals("129", vals.toString()); //bits=259
}

/** Test of class {@code Values(Iterable<Values> valueses)} constructor. */
@Test
public void testConstructor_iterableValues() {
	LinkedList<Values> list = new LinkedList<>();
	list.add(new Values(1,2));
	list.add(new Values(2,9));
	list.add(new Values(1,9));
	Values vals = new Values(list); // Iterable<Values>
	assertEquals(3, vals.size);
	assertEquals(259, vals.bits); //1,2,9
}

// ---------------- methods ----------------

/**
 * Test of copyFrom method, of class Values.
 */
@Test
public void testCopyFrom() {
	final String testName = testName(this);
	Values src = new Values(259, false); //1,2,9
	Values vals = Values.none(); // ie empty
	vals.copyFrom(src);
	assertEquals(3, vals.size);
	assertEquals(259, vals.bits); //1,2,9
}

/**
 * Test of fill method, of class Values.
 */
@Test
public void testFill() {
	final String testName = testName(this);
	Values vals = Values.none(); // empty
	vals.fill();
	assertEquals(9, vals.size);
	assertEquals(511, vals.bits); // (1<<10) - 1
	assertEquals("123456789", vals.toString());
}

/**
 * Test of clear method, of class Values.
 */
@Test
public void testClear() {
	final String testName = testName(this);
	Values vals = Values.all(); // 123456789
	vals.clear();
	assertEquals(0, vals.size);
	assertEquals(0, vals.bits);
}

/**
 * Test of clear method, of class Values.
 */
@Test
public void testClear_int() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	vals.remove(2);
	assertEquals(2, vals.size);
	assertEquals(257, vals.bits); //1,9
}

/**
 * Test of clear method, of class Values.
 */
@Test
public void testClear_Values() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	vals.remove(new Values(2,9));
	assertEquals(1, vals.size);
	assertEquals(1, vals.bits); //1
}

/**
 * Test of clearBits method, of class Values.
 */
@Test
public void testClearBits() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	vals.removeBits(3);
	assertEquals(1, vals.size);
	assertEquals(256, vals.bits); //9
}

/**
 * Test of add method, of class Values.
 */
@Test
public void testAdd_int() {
	final String testName = testName(this);
	Values vals = new Values(1,9);
	vals.add(2);
	assertEquals(3, vals.size);
	assertEquals(259, vals.bits); //1,2,9
}

/**
 * Test of add method, of class Values.
 */
@Test
public void testAdd_intArr() {
	final String testName = testName(this);
	Values vals = new Values(9);
	vals.add(1,2);
	assertEquals(3, vals.size);
	assertEquals(259, vals.bits); //1,2,9
}

/**
 * Test of set method, of class Values.
 */
@Test
public void testSet() {
	final String testName = testName(this);
	Values vals = Values.none();
	vals.set("129");
	assertEquals(3, vals.size);
	assertEquals(259, vals.bits); //1,2,9
}

/**
 * Test of add method, of class Values.
 */
@Test
public void testAdd_Values() {
	final String testName = testName(this);
	Values vals = Values.none();
	vals.add(1);
	vals.add(2);
	vals.add(9);
	assertEquals(3, vals.size);
	assertEquals(259, vals.bits); //1,2,9
}

/**
 * Test of visit method, of class Values.
 */
@Test
public void testFirstVisit() {
	final String testName = testName(this);
	Values vals = new Values(1,2);
	// note that vals IS modified by visit
	assertEquals(true, vals.visit(9));
	assertEquals(false, vals.visit(1));
	assertEquals(false, vals.visit(2));
	assertEquals(false, vals.visit(9)); // the clincher
}

/**
 * Test of intersect method, of class Values.
 */
@Test
public void testIntersect_int() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	// note that vals is NOT modified by intersect
	assertEquals(2, vals.intersect(2).bits);
	assertEquals(256, vals.intersect(9).bits);
	assertEquals(0, vals.intersect(3).bits);
}

/**
 * Test of intersect method, of class Values.
 */
@Test
public void testIntersect_Values() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	// note that vals is NOT modified by intersect
	assertEquals(2, vals.intersect(new Values(2)).bits);
	assertEquals(256, vals.intersect(new Values(9)).bits);
	assertEquals(0, vals.intersect(new Values(3)).bits);
	assertEquals(2, vals.intersect(new Values(2,3)).bits);
}

/**
 * Test of intersect method, of class Values.
 */
@Test
public void testIntersect_Values_Values() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	// note that vals is NOT modified by intersect
	assertEquals(2, vals.intersect(newValues(1,2,9), newValues(2,3,4,5,6,7,8)).bits);
	assertEquals(0, vals.intersect(newValues(1,2,3,4,5,6,7,8,9), new Values()).bits);
	assertEquals(0, vals.intersect(newValues(1,2), newValues(3,4,5,6,7,8,9)).bits);
}

/**
 * Test of minus method, of class Values.
 */
@Test
public void testMinus_int() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	// note that vals is NOT modified by minus
	assertEquals(3, vals.minus(9).bits);
	assertEquals(256, vals.minus(1).minus(2).bits);
	assertEquals(259, vals.minus(new Values()).bits); //1,2,9
	assertEquals(259, vals.minus(newValues(3,4,5,6,7,8)).bits); //1,2,9
	assertEquals(0, new Values().minus(new Values(511,F)).bits); //1..9
}

/**
 * Test of minus method, of class Values.
 */
@Test
public void testMinus_Values() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	// note that vals is NOT modified by minus
	assertEquals(0, vals.minus(newValues(1,2,9)).bits);
	assertEquals(2, vals.minus(new Values(1,9)).bits);
	assertEquals(259, vals.minus(new Values()).bits); //1,2,9
	assertEquals(259, vals.minus(newValues(3,4,5,6,7,8)).bits); //1,2,9
	assertEquals(0, new Values().minus(new Values(511,F)).bits); //1..9
}

/**
 * Test of minus method, of class Values.
 */
@Test
public void testMinus_Values_int() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	// note that vals is NOT modified by minus
	assertEquals(2, vals.minus(newValues(4,5,9), 1).bits); //2
	assertEquals(0, vals.minus(newValues(1,2,6), 9).bits);
	assertEquals(259, vals.minus(newValues(3,4,5,6,7), 8).bits); //1,2,9
	assertEquals(256, vals.minus(new Values(1), 2).bits); //9
	assertEquals(3, vals.minus(new Values(), 9).bits); //1,2
}

/**
 * Test of minus method, of class Values.
 */
@Test
public void testMinus_int_int() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	// note that vals is NOT modified by minus
	assertEquals(256, vals.minus(1,2).bits); //9
	assertEquals(259, vals.minus(3,4).minus(5,6).minus(7,8).bits); //1,2,9
	assertEquals(0, vals.minus(1,2).minus(7,9).bits);
}

/**
 * Test of plus method, of class Values.
 */
@Test
public void testPlus_Values() {
	final String testName = testName(this);
	Values vals = new Values(1,2);
	// note that vals is NOT modified by plus
	assertEquals(259, vals.plus(new Values(2,9)).bits); //1,2,9
	assertEquals(3, vals.plus(new Values(1,2)).bits); //1,2
	assertEquals(3, vals.plus(new Values()).bits); //1,2
}

/**
 * Test of plus method, of class Values.
 */
@Test
public void testPlus_Values_Values() {
	final String testName = testName(this);
	Values vals = new Values(1,2);
	// note that vals is NOT modified by plus
	assertEquals(259, vals.plus(new Values(1,2), new Values(2,9)).bits); //1,2,9
	assertEquals(3, vals.plus(new Values(1,2), new Values(1,2)).bits); //1,2
	assertEquals(3, vals.plus(new Values(), new Values()).bits); //1,2
}

/**
 * Test of plusClear method, of class Values.
 */
@Test
public void testPlusClear_3args() {
	final String testName = testName(this);
	Values vals = new Values(1,2);
	// note that vals is NOT modified by plusClear
	assertEquals(257, vals.plusClear(new Values(1,9), 2, 8).bits); //1,9
	assertEquals(1, vals.plusClear(new Values(1,9), 2, 9).bits); //1
	assertEquals(1, vals.plusClear(new Values(1,9), 2, 9).bits); //1
	assertEquals(256, vals.plusClear(new Values(2,9), 1, 2).bits); //9
}

/**
 * Test of get method, of class Values.
 */
@Test
public void testGet() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	assertEquals(T, vals.contains(1));
	assertEquals(T, vals.contains(2));
	assertEquals(F, vals.contains(3));
	assertEquals(F, vals.contains(8));
	assertEquals(T, vals.contains(9));
}

/**
 * Test of no method, of class Values.
 */
@Test
public void testNo() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	assertEquals(F, vals.no(1)); // double negative
	assertEquals(F, vals.no(2)); // double negative
	assertEquals(T, vals.no(3));
	assertEquals(T, vals.no(8));
	assertEquals(F, vals.no(9)); // double negative
}

/**
 * Test of containsAll method, of class Values.
 */
@Test
public void testContainsAll() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	assertEquals(T, vals.containsAll(new Values(1)));
	assertEquals(T, vals.containsAll(new Values(1,2)));
	assertEquals(T, vals.containsAll(newValues(1,2,9)));
	assertEquals(F, vals.containsAll(newValues(1,2,6,9)));
	assertEquals(F, vals.containsAll(new Values())); // that's a bit odd!
}

/**
 * Test of isSubsetOf method, of class Values.
 */
@Test
public void testIsSubsetOf_Values() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	assertEquals(T, vals.isSubsetOf(newValues(1,2,9)));
	assertEquals(T, vals.isSubsetOf(newValues(1,2,6,9)));
	assertEquals(F, vals.isSubsetOf(newValues(1,4,6,9)));
}

/**
 * Test of isEmpty method, of class Values.
 */
@Test
public void testIsEmpty() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	assertEquals(F, vals.isEmpty());
	vals = new Values();
	assertEquals(T, vals.isEmpty());
}

/**
 * Test of first method, of class Values.
 */
@Test
public void testFirst() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	assertEquals(1, vals.first());
	vals = new Values();
	assertEquals(-1, vals.first());
}

/**
 * Test of last method, of class Values.
 */
@Test
public void testLast() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	assertEquals(9, vals.last());
	vals = new Values();
	assertEquals(-1, vals.last());
}

/**
 * Test of next method, of class Values.
 */
@Test
public void testNext() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	assertEquals(1, vals.next(1));
	assertEquals(2, vals.next(2));
	assertEquals(9, vals.next(3));
	assertEquals(9, vals.next(9));
	assertEquals(-1, vals.next(10));
	vals = new Values();
	assertEquals(-1, vals.next(1));
}

/**
 * Test of next0 method, of class Values.
 */
@Test
public void testNext0() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	assertEquals(3, vals.next0(1));
	assertEquals(4, vals.next0(4));
	assertEquals(8, vals.next0(8));
	assertEquals(-1, vals.next0(9));
	vals = new Values(1,2);
	assertEquals(-1, vals.next(9));
	vals = new Values();
	assertEquals(-1, vals.next(9));
}

/**
 * Test of toArray method, of class Values.
 */
@Test
public void testToArray_intArr() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	int expSize = 3;
	int[] expArray = new int[]{1,2,9,0,0,0,0,0,0};
	int[] actArray = new int[9];
	int n = vals.toArray(actArray);
	assertEquals(expSize, n);
	assertArrayEquals(expArray, actArray);
	// now remove 2 elements
	vals.remove(2);
	vals.remove(9);
	expSize = 1;
	// either use the returned n OR respect the null terminator.
	// note the dirty 9      v
	expArray = new int[]{1,0,9,0,0,0,0,0,0};
	n = vals.toArray(actArray);
	assertEquals(expSize, n);
	assertArrayEquals(expArray, actArray);
}

/**
 * Test of toArray method, of class Values.
 */
@Test
public void testToArray() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	assertArrayEquals(new int[]{1,2,9}, vals.toArrayNew());
	// now remove 2 elements
	vals.remove(2);
	vals.remove(9);
	assertArrayEquals(new int[]{1}, vals.toArrayNew());
}

/**
 * Test of toBinaryString method, of class Values.
 */
@Test
public void testToBinaryString() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	assertEquals("100000011", vals.toBinaryString());
	vals.remove(9);
	assertEquals("000000011", vals.toBinaryString()); // NOTE the leading zeros
}

/**
 * Test of appendTo method, of class Values.
 */
@Test
public void testAppendTo() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	StringBuilder sb = new StringBuilder();
	assertEquals("129", Values.appendTo(sb, vals.bits).toString());
}

/**
 * Test of toString method, of class Values.
 */
@Test
public void testToString_int() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	assertEquals("129", Values.toString(vals.bits));
}

/**
 * Test of toString method, of class Values.
 */
@Test
public void testToString() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	assertEquals("129", vals.toString());
}

/**
 * Test of equals method, of class Values.
 */
@Test
public void testEquals_Object() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	Object that = new Object();
	assertEquals(false, vals.equals(that));
	that = new Values(vals);
	assertEquals(true, vals.equals(that));
	((Values)that).remove(9);
	assertEquals(false, vals.equals(that));
}

/**
 * Test of equals method, of class Values.
 */
@Test
public void testEquals_Values() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	Values that = new Values();
	assertEquals(false, vals.equals(that));
	that.copyFrom(vals);
	assertEquals(true, vals.equals(that));
	that.remove(9);
	assertEquals(false, vals.equals(that));
}

/**
 * Test of hashCode method, of class Values.
 */
@Test
public void testHashCode() {
	final String testName = testName(this);
	Values vals = newValues(1,2,9);
	assertEquals(259, vals.hashCode());
	vals.remove(9);
	assertEquals(3, vals.hashCode());
	// note that this breaks the hashCode contract, I forget what they call
	// it but it's supposed to return the same value for the same object
	// reliably, which implies that only non-volatile (ie final) identity
	// fields may be used in the hash and therefore in equals... but I don't
	// have any of them, nor do I want them, so this will have to do.
}

}