/*
 * Project: Sudoku Explainer
 * Copyright (C) 2006-2007 Nicolas Juillerat
 * Copyright (C) 2013-2020 Keith Corlett
 * Available under the terms of the Lesser General Public License (LGPL)
 */
package diuf.sudoku.solver;

import static diuf.sudoku.TestHelp.*;

import org.junit.Test;
import diuf.sudoku.Settings;
import diuf.sudoku.Tech;
import java.util.Arrays;
import java.util.EnumSet;
import java.util.HashSet;


/**
 * Tests the LogicalAnalyser class, an AHinter.
 * @author Keith Corlett 2018 Mar
 */
public final class LogicalAnalyserTest {

private static final String PUZZLE
="7.....4...2..7..8...3..8..9...5..3...6..2..9...1..7..6...3..9...3..4..6...9..1..5";
//1#top1465.d5.mt is one of my 3 puzzle which requires DynamicPlus to solve.

private static final boolean REPRODUCE = REPRODUCE_ALL_TESTS || false;

/** Test of LogicalAnalyser.getHints. */
@Test
public void test() {
	final String testName = testName(this);

	// ensure that the required hinters are enabled, by turning them on before
	// we create the LogicalSolver, I'm doing this because I'm well sick of
	// my effing-around with Selected Techniques in the GUI stuffing-up the
	// test-case results, just because I forgot to re-enable things.
	// We also need to do de-hack Aligned456Exclusion hinters.
	String requiredNames
="TooFewClues, TooFewValues"													// puzzle validators
+", NoDoubleValues, NoMissingMaybes, NoHomelessValues, SingleSolution"			// grid checkers
+", NakedSingle, HiddenSingle, Locking"											// direct hinters
+", DirectNakedPair, DirectHiddenPair, DirectNakedTriple, DirectHiddenTriple"	// indirect direct
+", NakedPair, NakedTriple, NakedQuad, NakedPent"								// naked hinters
+", HiddenPair, HiddenTriple, HiddenQuad, HiddenPent"							// hidden hinters
+", Swampfish, Swordfish, Jellyfish, XY_Wing, XYZ_Wing"
+", W_Wing, Skyscraper, TwoStringKite, EmptyRectangle"			// HoDoKu
+", Coloring, WXYZ_Wing, VWXYZ_Wing, UVWXYZ_Wing, TUVWXYZ_Wing"	// HoDoKu
+", FinnedSwampfish, FinnedSwordfish, FinnedJellyfish"			// HoDoKu (includes Sashimi)
+", FrankenSwampfish, FrankenSwordfish, FrankenJellyfish"		// HoDoKu
+", URT, ALS_XZ, ALS_Wing, ALS_Chain, DeathBlossom, SueDeCoq"	// HoDoKu ALSs
+", MutantSwampfish, MutantSwordfish, MutantJellyfish"			// HoDoKu (too slow)
+", KrakenSwampfish, KrakenSwordfish, KrakenJellyfish"			// HoDoKu Fish + Chaining
+", AlignedPair, AlignedTriple, AlignedQuad"									// fast enough
+", AlignedPent, AlignedHex, AlignedSept, AlignedOct, AlignedNona, AlignedDec"	// slow -> A7+E too slow correct
+", UnaryChain, NishioChain, MultipleChain, DynamicChain, DynamicPlus"			// fast enough
+", NestedPlus"																	// the catch all
;
	EnumSet<Tech> required = Tech.forNames(new HashSet<>(Arrays.asList(requiredNames.split(", "))));
	EnumSet<Tech> before = Settings.THE.setWantedTechniques(required);
	boolean preA4E = Settings.THE.set(Settings.isa4ehacked, false);
	boolean preA5E = Settings.THE.set(Settings.isa5ehacked, false);
	boolean preA6E = Settings.THE.set(Settings.isa6ehacked, true);
	boolean preA7E = Settings.THE.set(Settings.isa7ehacked, true);
	boolean preA8E = Settings.THE.set(Settings.isa8ehacked, true);
	boolean preA9E = Settings.THE.set(Settings.isa9ehacked, true);
	boolean preA10E = Settings.THE.set(Settings.isa10ehacked, true);
	Settings.THE.save();
	try {
		// create the LogicalSolver AFTER setting the wanted techniques.
		LogicalSolver solver = LogicalSolverFactory.get();
		// "manually" activate the IHintNumberActivatableHinter's. Sigh.
		solver.activate_4JUnit(true);
		// create my hinter
		LogicalAnalyser hinter = new LogicalAnalyser(solver); // implements IPreparer with a no-op
		// make the analyser print hints
		hinter.isNoisy = true;
		// get the hints from the hinter
		// MUST use the getHints which prepares the grid (once it's loaded),
		// otherwise we get an NPE from Idx$CellsMaybesVisitor.visit; so I
		// had to make LogicalAnalser implement IPreparer with a no-op. sigh.
		// This is (AFAIK) the ONLY case were we need to prepare the logical
		// solver for a bloody hinter that doesn't require preparation. sigh.
		getHints(PUZZLE, hinter, true);
		// get the hint back out of the IAccumulator
		AHint hint = accu.getHint();
		// check the hint against a stored HTML
		assertHintEqualsStoredHintsHtml(this, hint, testName+".html", REPRODUCE);
	} finally {
		Settings.THE.justSetWantedTechniques(before);
		Settings.THE.set(Settings.isa4ehacked, preA4E);
		Settings.THE.set(Settings.isa5ehacked, preA5E);
		Settings.THE.set(Settings.isa6ehacked, preA6E);
		Settings.THE.set(Settings.isa7ehacked, preA7E);
		Settings.THE.set(Settings.isa8ehacked, preA8E);
		Settings.THE.set(Settings.isa9ehacked, preA9E);
		Settings.THE.set(Settings.isa10ehacked, preA10E);
		Settings.THE.save();
	}
}

}
