/*
 * Project: Sudoku Explainer
 * Copyright (C) 2006-2007 Nicolas Juillerat
 * Copyright (C) 2013-2020 Keith Corlett
 * Available under the terms of the Lesser General Public License (LGPL)
 */
package diuf.sudoku.solver;

import diuf.sudoku.Grid;
import static diuf.sudoku.TestHelp.testName;
import diuf.sudoku.solver.accu.HintsAccumulator;
import diuf.sudoku.solver.accu.IAccumulator;
import java.util.LinkedList;
import static org.junit.Assert.assertEquals;
import org.junit.Test;

/**
 * Test some methods of the LogicalSolver class. Please note that other
 * existing test-cases exercise many of LogicalSolvers methods, and the net
 * result is far from complete. Stuff like getFirstHint isn't tested because
 * it's used in the GUI, which I test manually because (from experience)
 * writing test-cases for a bloody GUI is more of a problem than a cure; it
 * just makes every single GUI-change take atleast twice as long because you
 * have to find ALL test-cases which rely on whatever you've changed and modify
 * them also; and it gets worse: the automated-testing-framework we used had no
 * concept of a method, so you couldn't create a clickOK() method, instead the
 * exact coordinates of the click are recorded in the "macro", so you have to
 * write a macro-parses to find all clicks with x between x1 and x2 and y
 * between y1 and y2... so the whole fucking thing was a nightmare, so we 
 * persisted with it for a while, trying to get the bastard to work cleanly 
 * enough to not have our GUI concreted-in by "the change enabler"; and then we
 * binned the whole mess.
 * 
 * @author Keith Corlett 2019 DEC
 */
public class LogicalSolverTest {


private static final String HIDDEN_PAIR
="...6....4..53.9..1....8.73..1...6...6.3..8...7..91....2..........7...58..9....6.3";
//8#C:\Users\User\Documents\SodukuPuzzles\MagicTour\top1465.d3.mt
//first hint is a Hidden Pair: E7, E8: 6, 9 in col E (E4+3 E7-3457, E8-234)

// I found this puzzle file using testFindEasyPuzzles:
// singles: C:\Users\User\Documents\SodukuPuzzles\Dell\050E.txt
private static final String SOLVES_WITH_SINGLES
="...9.2...3..71...84....5192..265..49..3.7.5..69..812..7841....59...27..6...3.8...";
private static final String SOLVED_WITH_SINGLES
="518962734329714658467835192172653849843279561695481273784196325931527486256348917";
private static final String NAKED_SINGLE_WITH_CONSEQUENCES =
 "Naked Single: D3=8 (D3+8), E3=3, D6=4, F4=3, D5=2, F5=9, I5=1, H5=6"
+", G4=8, A4=1, B4=7, A5=8, B5=4, C6=5, C8=1, C9=6, B8=3, D8=5, F7=6"
+", E7=9, E9=4, I9=7, G7=3, H7=2, G8=4, H8=8, G9=9, H9=1, H2=5, G2=6"
+", G1=7, H1=3, I1=4, A1=5, C1=8, B2=2, C2=9, B3=6, B1=1, C3=7, E1=6"
+", F2=4, B9=5, A9=2, I6=3, H6=7";

// I found this puzzle in C:\Users\User\Documents\NetBeansProjects\DiufSudoku\logs\top1465.d3.134.verbose.log:
// 18#C:\Users\User\Documents\NetBeansProjects\DiufSudoku\top1465.d3.mt
// it's solve starts with 4 naked and hidden singles
private static final String PARTIALLY_SOLVES_WITH_SINGLES
="..54......4..8.7.......3..5..3.26.891...........3.7.2...9.4...3.......6..2.5....8";
private static final String PARTIALLY_SOLVED_WITH_SINGLES
="..54......4..857.......3..5..3126.891.28.4......3.7.2...9.4...3.......6..2.5....8";

/**
 * Test Grid.solveWithSingles() method
 * and Grid.load(String) with a 1 liner
 * and the equals(Object) method.
 */
@Test
public void testSolveWithSingles1() {
	final String testName = testName(this);
	Grid solution;
	String expected, actual;

	Grid grid = new Grid();
	LogicalSolver solver = LogicalSolverFactory.get();

	// NOTE: these tests are all in one because we don't want to create a fresh
	// grid for each test-case, ie we're exercising the grids ability to keep
	// itself "clean", ie that the indexes are maintained in all circumstances;
	// including loading a new puzzle into a "used" grid.
	// NOTE: We're also testing solveWithSingles with a null stringBuilder.

	// 1. The simplest hint in this grid is a Hidden Pair, there are no naked
	//    or hidden singles, so wantPartial=true returns the grid unchanged.
	grid.load(HIDDEN_PAIR);
	solution = solver.solveWithSingles(grid, null)
			? solver.singlesSolution
			: grid;
	assertEquals(grid, solution);

	// 2. This puzzle solves with singles, so check AppliedHintsSummaryHint
	grid.load(SOLVES_WITH_SINGLES);
	LinkedList<AHint> hintsList = new LinkedList<>();
	IAccumulator accu = new HintsAccumulator(hintsList);
	boolean result = solver.solveWithSingles(grid, accu);
	assertEquals(true, result);
	solution = result ? solver.singlesSolution : null;
	assert solution != null;
	assertEquals(SOLVED_WITH_SINGLES, solution.toString());
	expected = NAKED_SINGLE_WITH_CONSEQUENCES;
	actual = hintsList.getFirst().toString();
	assertEquals(expected, actual);

	// 3a. This puzzle partially solves with singles, so:
	//     wantPartial=false gets null
	grid.load(PARTIALLY_SOLVES_WITH_SINGLES);
	solution = solver.solveWithSingles(grid, null)
			? solver.singlesSolution
			: null;
	assertEquals(null, solution);

	// 3b. This puzzle partially solves with singles, so:
	//     wantPartial=true gets the partially solved grid.
	grid.load(PARTIALLY_SOLVES_WITH_SINGLES);
	solver.solveWithSingles(grid, null);
	// get partially solved grid, regardless of solveWithSingles's return value
	expected = PARTIALLY_SOLVED_WITH_SINGLES;
	solution = solver.singlesSolution;
	actual = solution.toShortString();
	if ( !expected.equals(actual) ) {
		System.err.format("EXPECTED : %s\n", expected);
		System.err.format("ACTUAL   : %s\n", actual);
	}
	assertEquals(expected, actual);
	
}

}
