/*
 * Project: Sudoku Explainer
 * Copyright (C) 2006-2007 Nicolas Juillerat
 * Copyright (C) 2013-2020 Keith Corlett
 * Available under the terms of the Lesser General Public License (LGPL)
 */
package diuf.sudoku.solver.checks;

import static diuf.sudoku.TestHelp.*;
import static org.junit.Assert.*;
import diuf.sudoku.Grid;
import diuf.sudoku.solver.AHint;
import diuf.sudoku.solver.hinters.AHinter;
import java.util.Random;
import org.junit.Test;

/**
 * Tests the RecursiveAnalyser class, an AWarningHinter.
 * @author Keith Corlett 2018 Mar
 */
public final class RecursiveAnalyserTest {

private static final boolean REPRODUCE = REPRODUCE_ALL_TESTS;

private static final String PUZZLE =
"....41...9..5...81.7...96.....4.....3..6..14..8.1....3......83579.........2....1.";
//1465#C:\Users\User\Documents\SodukuPuzzles\MagicTour\top1465.d3.mt

/** Test RecursiveAnalyser.getHints: Single solution ==> No hint */
@Test
public void testOneSolution() {
	final String testName = testName(this);
	AHinter hinter = new RecursiveAnalyser(); // SolutionMode.IGNORED
	// Invalid grid: Single solution ==> No hint.
	getHints(PUZZLE, hinter);
	AHint hint = accu.getHint();
	assertEquals(null, hint);
}


private static final String TWO_SOLUTIONS =
// modified 1465#C:\Users\User\Documents\SodukuPuzzles\MagicTour\top1465.d3.mt
"....41...9..5...81.7...96.....4........6..14..8.1....3......83579.........2....1.";
//                                   3 is missing

/** Test RecursiveAnalyser.getHints: Invalid grid: Two solutions */
@Test
public void testTwoSolutions() {
	final String testName = testName(this);
	// Note that this must be a RecursiveAnalyser, not just an AHinter in
	// order to access the unusual fields and methods.
	RecursiveAnalyser hinter = new RecursiveAnalyser(); // SolutionMode.IGNORED
	getHints(TWO_SOLUTIONS, hinter);
	AHint hint = accu.getHint();
	if ( !(hint instanceof MultipleSolutionsHint) )
		fail("WTF it's a "+hint.getClass().getSimpleName());
	MultipleSolutionsHint msHint = (MultipleSolutionsHint)hint;
	if ( msHint.getSolution1().equals(msHint.getSolution2()) )
		fail("WTF solutions equal: "+NL+
				msHint.getSolution1()+NL+
				msHint.getSolution2());
	// No point testing the HTML because it's just a pro-forma, ie it doesn't
	// contain any details specific to the actual use-case.
}


private static final String NO_SOLUTION =
".5.24...3.6.....2.8.......94..812.....5.6......83.5.....6...1..3...792.5...1..7.." + NL
+"179,,179,,,1678,68,1678,,179,,13479,579,3589,1378,458,,1478,,12347,12347,567,35,1367,456,14567,,,379,379,,,,3569,35679,67,1279,12379,,479,,47,3489,134789,12478,12679,1279,,,9,,469,14679,12467,2579,24789,,45,2358,348,,3489,48,,148,14,46,,,,468,,259,2489,249,,2358,3468,,34689,468";
// modified 741#C:\Users\User\Documents\SodukuPuzzles\MagicTour\top1465.d3.mt (E5+6)
// This Sudoku has <b>no solution</b>.

/** Test RecursiveAnalyser.getHints. */
@Test
public void testNoSolution() {
	final String testName = testName(this);
	RecursiveAnalyser hinter = new RecursiveAnalyser(); // SolutionMode.IGNORED
	// Invalid grid: No solution.
	// NB: You can't use getHints here coz it's "pre-check" of the loaded grid
	// always fails because the grid is invalid. So we just DIY load it.
	HINTS.clear(); // for when get hints is called repeatedly
	grid = new Grid();
	grid.load(NO_SOLUTION);
	if ( !grid.isMaybesLoaded ) // NO_SOLUTION is multiline (with maybes)
		fail("FAILED TO LOAD MAYBES:\n"+NO_SOLUTION);
	// nb: accu is a HintsAccumulator which populates HINTS
	if ( !hinter.findHints(grid, accu) )
		System.out.println("WARNING: "+hinter+" found nothing.");

	AHint hint = accu.getHint();
	if ( !(hint instanceof WarningHint) )
		fail("OMG it's a "+hint.getClass().getSimpleName());
	WarningHint wHint = (WarningHint)hint;
	assertEquals("No solution", wHint.toString());
}

///** Test RecursiveAnalyser.getHints when SolutionMode.IGNORED. */
//@Test
//public void testSolutionIgnored() {
//	final String testName = testName(this);
//	RecursiveAnalyser hinter = new RecursiveAnalyser(); // SolutionMode.IGNORED
//	// SolutionMode.IGNORED
//	getHints(PUZZLE, hinter);
//	AHint hint = accu.getHint();
//	assertEquals(null, hint); // because SolutionMode.IGNORED
//	assertEquals(null, hinter.solutionHint); // because SolutionMode.IGNORED
//}
//
///** Test RecursiveAnalyser.getHints when SolutionMode.STORE. */
//@Test
//public void testSolutionStored() {
//	final String testName = testName(this);
//	RecursiveAnalyser hinter = new RecursiveAnalyser(); // SolutionMode.STORED
//	getHints(PUZZLE, hinter);
//	AHint hint = accu.getHint();
//	assertEquals(null, hint); // because SolutionMode.STORED
//	assertHintEqualsStoredHintsHtml(this, hinter.solutionHint, testName+".html", REPRODUCE);
//}

//KRC 2020-08-31 RecursiveAnalyser has changed to make SolutionMode and thereby
//this test "mostly irrelevant" and this test fails, so I have taken it out.
//private static final String SOLUTION =
//"865741392924536781173289654216493578357628149489157263641972835798315426532864917";
//
///** Test RecursiveAnalyser.getHints when SolutionMode.WANTED. */
//@Test
//public void testSolutionWanted() {
//	final String testName = testName(this);
//	RecursiveAnalyser recursiveAnalyser = new RecursiveAnalyser();
//
//	// tell the RecursiveAnalyser that I want the full SolutionHint.html
//	// which says that the solution is displayed as green potential values;
//	// otherwise it returns the plain "natural causes" SudokuSolved.html
//	recursiveAnalyser.wantSolutionHint = true; // SolutionMode.WANTED
//
//	getHints(PUZZLE, recursiveAnalyser);
//	AHint hint = accu.getHint();
//	if ( hint == null )
//		fail("WTF the hint is null!");
//	else if ( !(hint instanceof SolutionHint) )
//		fail("WTF it's a "+hint.getClass().getSimpleName());
//	SolutionHint solutionHint = (SolutionHint)hint;
//	Grid solution = solutionHint.getSolution();
//	//System.out.println(solution);
//	assertEquals(SOLUTION, solution.toString());
//	assertEquals(null, recursiveAnalyser.solutionHint);
//	assertHintEqualsStoredHintsHtml(this, hint, testName+".html", REPRODUCE);
//}

/**
 * Test of countSolutions method, of class RecursiveAnalyser.
 */
@Test
public void testCountSolutions() {
	final String testName = testName(this);
	RecursiveAnalyser analyser = new RecursiveAnalyser(); // SolutionMode.IGNORED
	// ONE solution
	grid = new Grid(PUZZLE);
	assertEquals(1, analyser.countSolutions(grid, true));
	// TWO solutions
	grid = new Grid(TWO_SOLUTIONS);
	assertEquals(2, analyser.countSolutions(grid, true));
}

/**
 * Test of buildRandomPuzzle method, of class RecursiveAnalyser.
 */
@Test
public void testBuildRandomPuzzle() {
	final String testName = testName(this);
	RecursiveAnalyser hinter = new RecursiveAnalyser(); // SolutionMode.IGNORED
	getHints(PUZZLE, hinter);
	Grid randomPuzzle = hinter.buildRandomPuzzle(Grid.RANDOM);
	int numSolutions = hinter.countSolutions(randomPuzzle, true);
	if (numSolutions != 1)
		fail(""+numSolutions+" solutions: "+randomPuzzle);
}

}
