/*
 * Project: Sudoku Explainer
 * Copyright (C) 2006-2007 Nicolas Juillerat
 * Copyright (C) 2013-2020 Keith Corlett
 * Available under the terms of the Lesser General Public License (LGPL)
 */
package diuf.sudoku.solver.hinters.als;

import static diuf.sudoku.TestHelp.*;
import static org.junit.Assert.*;
import org.junit.Test;
import diuf.sudoku.solver.AHint;

/**
 * Test the WWing hinter.
 * @author Keith Corlett 2020 Apr
 */
public final class DeathBlossomTest {

private static final boolean REPRODUCE = REPRODUCE_ALL_TESTS || false;

private static final DeathBlossom HINTER = new DeathBlossom();

/*
TEST DATA: 69#top1465.d5.mt is DeathBlossom central. AFAIK it contains the only
multi-cell, multi-value DeathBlossom in top1465; which is (closely related) the
only multiple-death-blossom-hints-in-a-grid.

.7...9..69..73....8.265.9..5........4...96....8.4.52......8.15...35....4.589.....
13,,145,128,124,,3458,12348,,,146,1456,,,248,458,1248,1258,,134,,,,14,,1347,137,,12369,1679,1238,127,12378,3467,134679,13789,,123,17,1238,,,3578,1378,13578,1367,,1679,,17,,,13679,1379,267,469,4679,23,,237,,,2379,127,19,,,1267,127,678,26789,,127,,,,12467,12347,367,2367,237
69#C:\Users\User\Documents\NetBeansProjects\DiufSudoku\top1465.d5.mt

DeathBlossom	Death Blossom: stem F3 in box 8 row 9 (E9-27)
DeathBlossom	Death Blossom: stem F3 in box 8 col E (E8-27, E9-27, F4-7)
DeathBlossom	Death Blossom: stem F3 in box 8 col E (E9-27)
DeathBlossom	Death Blossom: stem F3 in box 8 col E (E8-27, E9-27, F4-7)
DeathBlossom	Death Blossom: stem F3 in box 8 col E (E9-27)
DeathBlossom	Death Blossom: stem D7 in col F box 5 (E8-7, E9-7, F4-17)
DeathBlossom	Death Blossom: stem B8 in box 8 box 9 (F9-3, I7-3)
DeathBlossom	Death Blossom: stem B8 in box 8 box 9 (F9-3, I7-3)
DeathBlossom	Death Blossom: stem B8 in row 9 row 7 (F9-3, I7-3)

Note that the GUI, with finterHints switched-on (which is my normal) filters
out several of the hints listed above, which are distinct hints, on different
ALSs in the same regions, which produce the "same" hints.

*/
private static final String PUZZLE
=".7...9..69..73....8.265.9..5........4...96....8.4.52......8.15...35....4.589....." + NL
+"13,,145,128,124,,3458,12348,,,146,1456,,,248,458,1248,1258,,134,,,,14,,1347,137,,12369,1679,1238,127,12378,3467,134679,13789,,123,17,1238,,,3578,1378,13578,1367,,1679,,17,,,13679,1379,267,469,4679,23,,237,,,2379,127,19,,,1267,127,678,26789,,127,,,,12467,12347,367,2367,237";
//Death Blossom: stem F3 in box 8 col E (E8-27, E9-27, F4-7)
@Test
public void test1() {
	final String testName = testName(this);
	// nb: use the getHints which prepares the hinter.
	getHints(PUZZLE, HINTER);
	// find my hint (note that we need the full string) and even it isn't unique
	AHint hint = find(HINTS, "Death Blossom: stem F3 in box 8 col E (E8-27, E9-27, F4-7)");
////KEEP THESE: For use in creating test cases.
//System.out.format("aquaCells  : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getAquaCells(0)));
//System.out.format("oranges    : %s\n", hint.getOranges(0));
//System.out.format("greens     : %s\n", hint.getGreens(0));
//System.out.format("blues      : %s\n", hint.getBlues(grid, 0));
//System.out.format("reds       : %s\n", hint.getReds(0));
//System.out.format("redPots    : %s\n", hint.redPots);
//System.out.format("clueHtml   : %s\n", hint.getClueHtml(true));
//System.out.format("bases      : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getBases()));
//System.out.format("covers     : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getCovers()));
	checkClue(hint, "Look for a Death Blossom stem F3 in box 8 col E");
	assertEquals(null, hint.getAquaCells(0));
	assertEquals(newPots("F1-3"), hint.getGreens(0));
	assertEquals(null, hint.getOranges(0)); // oranges are now blue, blues are now yellow, oranges are now the 4th stem-cell-value/ALS, which don't exist in top1465; all because an orange background hides the brown link-arrows
	assertEquals(newPots("E1-4"), hint.getBlues(grid, 0));
	assertEquals(newPots("E8-27, E9-27, F4-7"), hint.getReds(0));
	assertEquals(newPots("E8-27, E9-27, F4-7"), hint.redPots);
	assertArrayEquals(regions("box 8, col E"), hint.getBasesArray());
	assertArrayEquals(null, hint.getCoversArray());
	assertHintEqualsStoredHintsHtml(this, hint, testName+".html", REPRODUCE);
}

}
