/*
 * Project: Sudoku Explainer
 * Copyright (C) 2006-2007 Nicolas Juillerat
 * Copyright (C) 2013-2020 Keith Corlett
 * Available under the terms of the Lesser General Public License (LGPL)
 */
package diuf.sudoku.solver.hinters.fish;

import static org.junit.Assert.*;
import static diuf.sudoku.TestHelp.*;
import org.junit.Test;
import diuf.sudoku.Tech;
import diuf.sudoku.solver.AHint;
import diuf.sudoku.solver.hinters.AHinter;


/**
 * Test the BasicFisherman hinter.
 * Tech=degree: Swampfish=2, Swordfish=3, Jellyfish=4
 * @author Keith Corlett 2018 Apr
 */
public final class BasicFishermanTest {

private static final boolean REPRODUCE = REPRODUCE_ALL_TESTS || false;

// -------------------------------- SWAMPFISH ---------------------------------

private static final String SWAMPFISH
="7.....4...2..7..8...3..8..9...5.63...6..2..9...1..7..6...3.29...3..4..6...9..1.35" + NL
+",89,68,26,13,59,,25,13,19,,56,149,,34,56,,13,14,145,,26,15,,267,27,,289,789,247,,189,,,14,2478,358,,457,18,,34,1578,,478,2359,589,,48,39,,258,245,,1568,1578,578,,568,,,147,478,158,,2578,789,,59,178,,278,2468,48,,78,68,,278,,";
//1#C:\Users\User\Documents\NetBeansProjects\DiufSudoku\top1465.d3.mt
//Swampfish: col C, col I and row 4, row 8 on 2 (A4-2)

/** Test Swampfish (degree == 2)... nee X-Wing */
@Test
public void testSwampfish() {
	final String testName = testName(this);
	AHinter hinter = new BasicFisherman(Tech.Swampfish);
	getHints(SWAMPFISH, hinter);
	AHint hint = find(HINTS, "Swampfish: col C, col I and row 4, row 8 on 2");
////KEEP THESE: For use in creating test cases.
//System.out.format("aquaCells  : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getAquaCells(0)));
//System.out.format("oranges    : %s\n", hint.getOranges(0));
//System.out.format("greens     : %s\n", hint.getGreens(0));
//System.out.format("reds       : %s\n", hint.getReds(0));
//System.out.format("redPots    : %s\n", hint.redPots);
//System.out.format("clueHtml   : %s\n", hint.getClueHtml(true));
//System.out.format("bases      : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getBases()));
//System.out.format("covers     : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getCovers()));
	// check that this is an indirect hint (has caused me some confusion)
	assertEquals(false, hint.getClass().getSimpleName().contains("Direct"));
	assertEquals(null, hint.cell);
	assertEquals(0, hint.value);
	checkClue(hint, "Look for a Swampfish on <b>2</b>");
	assertEquals(cellSet("C4, C8, I4, I8"), hint.getAquaCells(0));
	assertEquals(newPots("C4-2, C8-2, I4-2, I8-2"), hint.getGreens(0));
	assertEquals(newPots("A4-2"), hint.getReds(0));
	assertArrayEquals(regions("col C, col I"), hint.getBasesArray());
	assertArrayEquals(regions("row 4, row 8"), hint.getCoversArray());
	assertHintEqualsStoredHintsHtml(this, hint, testName+".html", REPRODUCE);
}


// -------------------------------- SWORDFISH ---------------------------------

private static final String SWORDFISH
="...3..5...5..1..3...7..4..12.....4...6..9......1..6..28..7..2..79..8..5...5..9..7" + NL
+"1469,1248,24689,,267,278,,246789,4689,469,,24689,2689,,278,6789,,4689,369,238,,25689,256,,689,2689,,,378,389,158,357,13578,,16789,35689,345,,348,12458,,123578,1378,178,358,3459,3478,,458,3457,,3789,789,,,134,346,,3456,135,,1469,3469,,,2346,1246,,123,136,,346,1346,1234,,1246,2346,,1368,1468,";
//7#C:\Users\User\Documents\NetBeansProjects\DiufSudoku\top1465.d3.mt
//Swordfish: row 2, row 5, row 8 and col C, col D, col F on 2 (C1-2, D3-2, D9-2, F1-2)
//Swordfish: col B, col E, col H and row 1, row 3, row 9 on 2 (C1-2, D3-2, D9-2, F1-2)

/** Test Swordfish 3 bases and 3 covers. */
@Test
public void testSwordfish() {
	final String testName = testName(this);
	AHinter hinter = new BasicFisherman(Tech.Swordfish);
	getHints(SWORDFISH, hinter);

	// HINT 1
	//System.out.println(accu.getHint().toString());
	AHint hint = find(HINTS, "Swordfish: row 2, row 5, row 8 and col C, col D, col F on 2");
////KEEP THESE: For use in creating test cases.
//System.out.format("clueHtml   : %s\n", hint.getClueHtml(true));
//System.out.format("aquaCells  : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getAquaCells(0)));
//System.out.format("oranges    : %s\n", hint.getOranges(0));
//System.out.format("greens     : %s\n", hint.getGreens(0));
//System.out.format("reds       : %s\n", hint.getReds(0));
//System.out.format("redPots    : %s\n", hint.redPots);
//System.out.format("bases      : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getBases()));
//System.out.format("covers     : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getCovers()));
	assertEquals(false, hint.getClass().getSimpleName().contains("Direct"));
	assertEquals(null, hint.cell);
	assertEquals(0, hint.value);
	checkClue(hint, "Look for a Swordfish on <b>2</b>");
	assertEquals(cellSet("C2, D2, F2, D5, F5, C8, D8, F8"), hint.getAquaCells(0));
	assertEquals(newPots("C2-2, C8-2, D2-2, D5-2, D8-2, F2-2, F5-2, F8-2"), hint.getGreens(0));
	assertEquals(newPots("C1-2, D3-2, D9-2, F1-2"), hint.getReds(0));
	assertArrayEquals(regions("row 2, row 5, row 8"), hint.getBasesArray());
	assertArrayEquals(regions("col C, col D, col F"), hint.getCoversArray());
	assertHintEqualsStoredHintsHtml(this, hint, testName+"_1.html", REPRODUCE, false);

	// HINT 2
	//System.out.println(accu.getHint().toString());
	hint = find(HINTS, "Swordfish: col B, col E, col H and row 1, row 3, row 9 on 2");
////KEEP THESE: For use in creating test cases.
//System.out.format("clueHtml   : %s\n", hint.getClueHtml(true));
//System.out.format("aquaCells  : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getAquaCells(0)));
//System.out.format("oranges    : %s\n", hint.getOranges(0));
//System.out.format("greens     : %s\n", hint.getGreens(0));
//System.out.format("reds       : %s\n", hint.getReds(0));
//System.out.format("redPots    : %s\n", hint.redPots);
//System.out.format("bases      : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getBases()));
//System.out.format("covers     : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getCovers()));
	checkClue(hint, "Look for a Swordfish on <b>2</b>");
	assertEquals(cellSet("B1, B3, B9, E1, E3, E9, H1, H3"), hint.getAquaCells(0));
	assertEquals(null, hint.getOranges(0));
	assertEquals(newPots("B1-2, B3-2, B9-2, E1-2, E3-2, E9-2, H1-2, H3-2"), hint.getGreens(0));
	assertEquals(newPots("C1-2, D3-2, D9-2, F1-2"), hint.getReds(0));
	assertArrayEquals(regions("col B, col E, col H"), hint.getBasesArray());
	assertArrayEquals(regions("row 1, row 3, row 9"), hint.getCoversArray());
	assertHintEqualsStoredHintsHtml(this, hint, testName+"_2.html", REPRODUCE);
}


// -------------------------------- JELLYFISH ---------------------------------

private static final String JELLYFISH
="581.94.3..2.6.59.......3......3.241713..4.26.......5.3.......2.....3.8...6.8.93.." + NL
+",,,27,,,67,,26,347,,347,,18,,,478,148,69,47,69,127,28,,17,458,2458,689,59,689,,56,,,,,,,57,59,,78,,,89,2467,47,2467,19,67,18,,89,,38,159,38,145,57,167,167,,469,2479,15,24579,1245,,167,,479,456,247,,2457,,12,,,457,145";
//138#C:\Users\User\Documents\NetBeansProjects\DiufSudoku\top1465.d3.mt
//Jellyfish: row 2, row 5, row 8, row 9 and col A, col C, col F, col H on 7 (A6-7, C6-7, F7-7)
//Jellyfish: col B, col D, col E, col G and row 1, row 3, row 6, row 7 on 7 (A6-7, C6-7, F7-7)

/** Test Jellyfish (degree == 4). */
@Test
public void testJellyfish() {
	final String testName = testName(this);
	AHinter hinter = new BasicFisherman(Tech.Jellyfish);
	getHints(JELLYFISH, hinter);

	// HINT 1
	AHint hint = find(HINTS, "Jellyfish: row 2, row 5, row 8, row 9 and col A, col C, col F, col H on 7");
////KEEP THESE: For use in creating test cases.
//System.out.format("clueHtml   : %s\n", hint.getClueHtml(true));
//System.out.format("aquaCells  : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getAquaCells(0)));
//System.out.format("oranges    : %s\n", hint.getOranges(0));
//System.out.format("greens     : %s\n", hint.getGreens(0));
//System.out.format("reds       : %s\n", hint.getReds(0));
//System.out.format("redPots    : %s\n", hint.redPots);
//System.out.format("bases      : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getBases()));
//System.out.format("covers     : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getCovers()));
	assertEquals(false, hint.getClass().getSimpleName().contains("Direct"));
	assertEquals(null, hint.cell);
	assertEquals(0, hint.value);
	checkClue(hint, "Look for a Jellyfish on <b>7</b>");
	assertEquals(cellSet("A2, C2, H2, C5, F5, A8, C8, F8, H8, A9, C9, H9"), hint.getAquaCells(0));
	assertEquals(newPots("A2-7, A8-7, A9-7, C2-7, C5-7, C8-7, C9-7, F5-7, F8-7, H2-7, H8-7, H9-7"), hint.getGreens(0));
	assertEquals(null, hint.getOranges(0));
	assertEquals(newPots("A6-7, C6-7, F7-7"), hint.getReds(0));
	assertEquals(newPots("A6-7, C6-7, F7-7"), hint.redPots);
	assertArrayEquals(regions("row 2, row 5, row 8, row 9"), hint.getBasesArray());
	assertArrayEquals(regions("col A, col C, col F, col H"), hint.getCoversArray());
	assertHintEqualsStoredHintsHtml(this, hint, testName+"_1.html", REPRODUCE, false);

	// HINT 2
	hint = find(HINTS, "Jellyfish: col B, col D, col E, col G and row 1, row 3, row 6, row 7 on 7");
////KEEP THESE: For use in creating test cases.
//System.out.format("clueHtml   : %s\n", hint.getClueHtml(true));
//System.out.format("aquaCells  : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getAquaCells(0)));
//System.out.format("oranges    : %s\n", hint.getOranges(0));
//System.out.format("greens     : %s\n", hint.getGreens(0));
//System.out.format("reds       : %s\n", hint.getReds(0));
//System.out.format("redPots    : %s\n", hint.redPots);
//System.out.format("bases      : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getBases()));
//System.out.format("covers     : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getCovers()));
	// check that this is an indirect hint (has caused me some confusion)
	assertEquals(false, hint.getClass().getSimpleName().contains("Direct"));
	assertEquals(null, hint.cell);
	assertEquals(0, hint.value);
	checkClue(hint, "Look for a Jellyfish on <b>7</b>");
	assertEquals(cellSet("B3, B6, D1, D3, E6, E7, G1, G3, G7"), hint.getAquaCells(0));
	assertEquals(newPots(""), hint.getGreens(0));
	assertEquals(null, hint.getOranges(0));
	assertEquals(newPots("A6-7, C6-7, F7-7"), hint.getReds(0));
	assertArrayEquals(regions("col B, col D, col E, col G"), hint.getBasesArray());
	assertArrayEquals(regions("row 1, row 3, row 6, row 7"), hint.getCoversArray());
	assertHintEqualsStoredHintsHtml(this, hint, testName+"_2.html", REPRODUCE || false);
}

}
