/*
 * Project: Sudoku Explainer
 * Copyright (C) 2006-2007 Nicolas Juillerat
 * Copyright (C) 2013-2020 Keith Corlett
 * Available under the terms of the Lesser General Public License (LGPL)
 */
package diuf.sudoku.solver.hinters.fish;

import static org.junit.Assert.*;
import static diuf.sudoku.TestHelp.*;
import org.junit.Test;
import diuf.sudoku.Tech;
import diuf.sudoku.Pots;
import diuf.sudoku.TestHelp;
import diuf.sudoku.solver.AHint;
import diuf.sudoku.solver.hinters.AHinter;
import diuf.sudoku.utils.Frmt;

/**
 * Test {@link ComplexFisherman#findHints}
 *
 * @author Keith Corlett 2020-10-09 Rewrite
 */
public final class ComplexFishermanTest {

static {
	// so that TestHelp.getHints rebuilds all the s__t in the grid.
	TestHelp.REBUILD_GRID = true;
}

private static final boolean REPRODUCE = REPRODUCE_ALL_TESTS || false;

// ============================================================================
//                                  FINNED FISH
// ============================================================================

private static final String FINNED_SWAMPFISH
="9....34.5........1...15986.7.953261...5...7...6287...9.94.251........2..2.86....3" + NL
+",18,16,27,68,,,27,,568,2578,67,247,468,4678,39,39,,34,2347,37,,,,,,27,,48,,,,,,,48,1348,1348,,49,1469,146,,2348,248,134,,,,,14,35,345,,36,,,37,,,,78,678,1356,1357,1367,3479,1489,1478,,4579,467,,157,,,149,147,59,4579,";
//C:/Users/User/Documents/SodukuPuzzles/HoDoKu/FinnedX-Wing_001.txt
//FinnedFisherman: Finned Swampfish: D1, H1, D7, H7 (fins I7) on 7 in 2 rows and 2 cols (H8-7, H9-7)
//ComplexFisherman: Finned Swampfish: row 1, row 7 and col D, col H on 7 (H8-7, H9-7)

/** Test Finned Swampfish (nee Finned X-Wing) (degree == 2). */
@Test
public void testFinnedSwampfish() {
	final String testName = testName(this);
	AHinter hinter = new ComplexFisherman(Tech.FinnedSwampfish);
	getHints(FINNED_SWAMPFISH, hinter);
	AHint hint = find(HINTS, "Finned Swampfish: row 1, row 7 and col D, col H on 7");
if ( REPRODUCE ) {
//KEEP THESE to maintain this testcase
System.out.format("hint        : %s\n", hint.toFullString());
System.out.format("aquaCells   : %s\n", Frmt.csv(hint.getAquaCells(0)));
System.out.format("oranges     : %s\n", hint.getOranges(0));
System.out.format("greens      : %s\n", hint.getGreens(0));
System.out.format("blue exoFin : %s\n", hint.getBlues(null, 0));
System.out.format("purple endo : %s\n", hint.getPurples());
System.out.format("yellow shark: %s\n", hint.getYellows());
System.out.format("reds        : %s\n", hint.getReds(0));
System.out.format("redPots     : %s\n", hint.redPots);
System.out.format("clueHtml    : %s\n", hint.getClueHtml(true));
System.out.format("bases       : %s\n", Frmt.csv(hint.getBases()));
System.out.format("covers      : %s\n", Frmt.csv(hint.getCovers()));
}
	assertEquals(null, hint.cell);
	assertEquals(0, hint.value);
	checkClue(hint, "Look for a Finned Swampfish on <b>7</b>");
	assertEquals(cellSet("D1, H1, D7, H7"), hint.getAquaCells(0));
	assertEquals(newPots("D1-7, H1-7, D7-7, H7-7"), hint.getGreens(0));
	assertEquals(newPots("I7-7"), hint.getBlues(null,0)); // exo-fins
	assertEquals(Pots.EMPTY, hint.getPurples()); // endo-fins
	assertEquals(null, hint.getYellows()); // sharks
	assertEquals(newPots("H8-7 H9-7"), hint.getReds(0));
	assertArrayEquals(regions("row 1, row 7"), hint.getBasesArray());
	assertArrayEquals(regions("col D, col H"), hint.getCoversArray());
	assertHintEqualsStoredHintsHtml(this, hint, testName+".html", REPRODUCE);
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

private static final String FINNED_SWORDFISH
="8.69.....5.1.6.9..3491.75...67..8.39.53....8.4986...5...23...75.34.1.8....5..23.4" + NL
+",27,,,2345,345,127,124,1237,,27,,248,,34,,24,2378,,,,,28,,,26,268,12,,,245,245,,124,,,12,,,247,2479,149,12467,,1267,,,,,237,13,127,,127,69,18,,,489,469,16,,,679,,,57,,569,,269,26,679,18,,78,789,,,169,";
//C:/Users/User/Documents/SodukuPuzzles/HoDoKu/FinnedSwordfish_001.txt
//Finned Swordfish: row 3, row 6, row 8 and col E, col H, col I on 2 (I5-2)

/** Test Finned Swordfish (degree == 3). */
@Test
public void testFinnedSwordfish() {
	final String testName = testName(this);
	AHinter hinter = new ComplexFisherman(Tech.FinnedSwordfish);
	getHints(FINNED_SWORDFISH, hinter);

	AHint hint = find(HINTS, "Finned Swordfish: row 3, row 6, row 8 and col E, col H, col I on 2");
if ( REPRODUCE ) {
//KEEP THESE to maintain this testcase
System.out.format("hint        : %s\n", hint.toFullString());
System.out.format("aquaCells   : %s\n", Frmt.csv(hint.getAquaCells(0)));
System.out.format("oranges     : %s\n", hint.getOranges(0));
System.out.format("greens      : %s\n", hint.getGreens(0));
System.out.format("blue exoFin : %s\n", hint.getBlues(null, 0));
System.out.format("purple endo : %s\n", hint.getPurples());
System.out.format("yellow shark: %s\n", hint.getYellows());
System.out.format("reds        : %s\n", hint.getReds(0));
System.out.format("redPots     : %s\n", hint.redPots);
System.out.format("clueHtml    : %s\n", hint.getClueHtml(true));
System.out.format("bases       : %s\n", Frmt.csv(hint.getBases()));
System.out.format("covers      : %s\n", Frmt.csv(hint.getCovers()));
}
	assertEquals(null, hint.cell);
	assertEquals(0, hint.value);
	checkClue(hint, "Look for a Finned Swordfish on <b>2</b>");
	assertEquals(cellSet("E3, H3, I3, H8, I8, E6, I6"), hint.getAquaCells(0));
	assertEquals(newPots("E3-2, H3-2, I3-2, H8-2, I8-2, E6-2, I6-2"), hint.getGreens(0));
	assertEquals(newPots("G6-2"), hint.getBlues(null,0)); // exo-fins
	assertEquals(Pots.EMPTY, hint.getPurples()); // endo-fins
	assertEquals(null, hint.getYellows()); // sharks
	assertEquals(newPots("I5-2"), hint.getReds(0));
	assertEquals(newPots("I5-2"), hint.redPots);
	assertArrayEquals(regions("row 3, row 6, row 8"), hint.getBasesArray());
	assertArrayEquals(regions("col E, col H, col I"), hint.getCoversArray());
	assertHintEqualsStoredHintsHtml(this, hint, testName+".html", REPRODUCE || false);
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/*
C:/Users/User/Documents/SodukuPuzzles/HoDoKu/FinnedJellyfish_001.txt
9..7.81.4.42..1..8.18....3..978...4.4...9.78282...759..89...46.2.46.981.6...84..9
,356,356,,235,,,25,,357,,,359,356,,69,57,,57,,,2459,2456,256,269,,567,135,,,,12356,2356,36,,136,,356,1356,135,,356,,,,,,36,34,1346,,,,136,135,,,235,12357,235,,,357,,357,,,357,,,,357,,357,135,135,,,23,257,
Finned Jellyfish: row 1, row 5, row 6, row 8 and col B, col C, col E, col I on 3 (E4-3)
*/
private static final String FINNED_JELLYFISH
="9..7.81.4.42..1..8.18....3..978...4.4...9.78282...759..89...46.2.46.981.6...84..9" + NL
+",356,356,,235,,,25,,357,,,359,356,,69,57,,57,,,2459,2456,256,269,,567,135,,,,12356,2356,36,,136,,356,1356,135,,356,,,,,,36,34,1346,,,,136,135,,,235,12357,235,,,357,,357,,,357,,,,357,,357,135,135,,,23,257,";
/** Test Finned Jellyfish (degree == 4). */
@Test
public void testFinnedJellyfish() {
	final String testName = testName(this);
	AHinter hinter = new ComplexFisherman(Tech.FinnedJellyfish);
	getHints(FINNED_JELLYFISH, hinter);
	AHint hint = find(HINTS, "Finned Jellyfish: row 1, row 5, row 6, row 8 and col B, col C, col E, col I on 3");
if ( REPRODUCE ) {
//KEEP THESE to maintain this testcase
System.out.format("hint        : %s\n", hint.toFullString());
System.out.format("aquaCells   : %s\n", Frmt.csv(hint.getAquaCells(0)));
System.out.format("oranges     : %s\n", hint.getOranges(0));
System.out.format("greens      : %s\n", hint.getGreens(0));
System.out.format("blue exoFin : %s\n", hint.getBlues(null, 0));
System.out.format("purple endo : %s\n", hint.getPurples());
System.out.format("yellow shark: %s\n", hint.getYellows());
System.out.format("reds        : %s\n", hint.getReds(0));
System.out.format("redPots     : %s\n", hint.redPots);
System.out.format("clueHtml    : %s\n", hint.getClueHtml(true));
System.out.format("bases       : %s\n", Frmt.csv(hint.getBases()));
System.out.format("covers      : %s\n", Frmt.csv(hint.getCovers()));
}
	assertEquals(null, hint.cell);
	assertEquals(0, hint.value);
	checkClue(hint, "Look for a Finned Jellyfish on <b>3</b>");
	assertEquals(cellSet("B1, C1, E1, B8, E8, I8, B5, C5, C6, E6, I6"), hint.getAquaCells(0));
	assertEquals(newPots("B1-3, C1-3, E1-3, B8-3, E8-3, I8-3, B5-3, C5-3, C6-3, E6-3, I6-3"), hint.getGreens(0));
	assertEquals(newPots("D5-3, F5-3, D6-3"), hint.getBlues(null,0)); // exo-fins
	assertEquals(Pots.EMPTY, hint.getPurples()); // endo-fins
	assertEquals(null, hint.getYellows()); // sharks
	assertEquals(newPots("E4-3"), hint.getReds(0));
	assertEquals(newPots("E4-3"), hint.redPots);
	assertArrayEquals(regions("row 1, row 5, row 6, row 8"), hint.getBasesArray());
	assertArrayEquals(regions("col B, col C, col E, col I"), hint.getCoversArray());
	assertHintEqualsStoredHintsHtml(this, hint, testName+".html", REPRODUCE);
}

// ============================================================================
//                                 SASHIMI FISH
// ============================================================================

private static final String SASHIMI_SWAMPFISH
="......3899.4..2561....9.724461927853859364172..2...496.97.1..485....89.7.....9..5" + NL
+"1267,127,56,14567,457,156,,,,,378,,78,378,,,,,136,138,3568,1568,,1356,,,,,,,,,,,,,,,,,,,,,,37,37,,158,58,15,,,,236,,,256,,356,26,,,,1234,36,246,34,,,13,,1236,12348,368,2467,347,,26,13,";
//C:/Users/User/Documents/SodukuPuzzles/HoDoKu/SashimiSwampfish_001.txt
//Sashimi Swampfish: col C, col F (fins C8, C9) and row 3, row 7 on 3 (A7-3)
// NOTE: SashimiSwampfish_001.txt and SashimiSwordfish_001.txt are exactly the
// same puzzle, which contains both; but be warned that "Sashimi Swampfish" is
// not the first hint found, so this test-case will not "work normally" in the
// GUI, you have to shift-F5 to see any Sashimi at all. This is hardly a shock
// when you consider that Sashimi is inherently degenerate; ie is composed of
// simpler hints. So you won't find m/any sashimi fish in LogicalSolverTester
// coz all simpler hints are found, in order.

/** Test Sashimi Swampfish (nee Sashimi X-Wing) (degree == 2). */
@Test
public void testSashimiSwampfish() {
	final String testName = testName(this);
	// nb: ComplexFisherman now also finds Sashimi in a Finned search,
	//     so there is no Tech.Sashimi.....fish any longer.
	AHinter hinter = new ComplexFisherman(Tech.FinnedSwampfish);
	getHints(SASHIMI_SWAMPFISH, hinter);
	AHint hint = find(HINTS, "Sashimi Swampfish: col C, col F (fins C8 C9) and row 3, row 7 on 3");
if ( REPRODUCE ) {
//KEEP THESE to maintain this testcase
System.out.format("hint        : %s\n", hint.toFullString());
System.out.format("aquaCells   : %s\n", Frmt.csv(hint.getAquaCells(0)));
System.out.format("oranges     : %s\n", hint.getOranges(0));
System.out.format("greens      : %s\n", hint.getGreens(0));
System.out.format("blue exoFin : %s\n", hint.getBlues(null, 0));
System.out.format("purple endo : %s\n", hint.getPurples());
System.out.format("yellow shark: %s\n", hint.getYellows());
System.out.format("reds        : %s\n", hint.getReds(0));
System.out.format("redPots     : %s\n", hint.redPots);
System.out.format("clueHtml    : %s\n", hint.getClueHtml(true));
System.out.format("bases       : %s\n", Frmt.csv(hint.getBases()));
System.out.format("covers      : %s\n", Frmt.csv(hint.getCovers()));
}
	assertEquals(null, hint.cell);
	assertEquals(0, hint.value);
	checkClue(hint, "Look for a Sashimi Swampfish on <b>3</b>");
	assertEquals(cellSet("F3, F7, C3"), hint.getAquaCells(0)); // 4th corner is the fins
	assertEquals(newPots("F3-3, F7-3, C3-3"), hint.getGreens(0));
	assertEquals(newPots("C8-3 C9-3"), hint.getBlues(null,0)); // the fins
	assertEquals(Pots.EMPTY, hint.getPurples());
	assertEquals(null, hint.getYellows());
	assertEquals(newPots("A7-3"), hint.getReds(0));
	assertEquals(newPots("A7-3"), hint.redPots);
	assertArrayEquals(regions("col C, col F"), hint.getBasesArray());
	assertArrayEquals(regions("row 3, row 7"), hint.getCoversArray());
	assertHintEqualsStoredHintsHtml(this, hint, testName+".html", REPRODUCE);
}

// ============================================================================
//                                 MUTANT FISH
// ============================================================================

/*
693#C:\Users\User\Documents\NetBeansProjects\DiufSudoku\top1465.d5.mt
1..47..6.8.....491.....1.73....1..39..982.1..5.19436..41....7.6..51.........6..1.
,39,23,,,89,258,,258,,2567,67,356,35,26,,,,269,2456,46,256,89,,28,,,267,468,4678,567,,567,258,,,367,367,,,,567,,45,457,,278,,,,,,28,278,,,238,235,3589,589,,258,,3679,23678,,,38,2478,39,248,248,2379,3789,278,257,,457,39,,2458
34   	     47,050,500	32	 143	  1	Mutant Swordfish              	Finned Mutant Swordfish: box 3, row 4, col C and row 1, row 9, col G on 2 (A9-2)
*/
private static final String MUTANT_SWORDFISH
="1..47..6.8.....491.....1.73....1..39..982.1..5.19436..41....7.6..51.........6..1." + NL
+",39,23,,,89,258,,258,,2567,67,356,35,26,,,,269,2456,46,256,89,,28,,,267,468,4678,567,,567,258,,,367,367,,,,567,,45,457,,278,,,,,,28,278,,,238,235,3589,589,,258,,3679,23678,,,38,2478,39,248,248,2379,3789,278,257,,457,39,,2458";
/** Test ComplexFisherman Tech.MutantSwordfish */
@Test
public void testMutantSwordfish() {
	final String testName = testName(this);
	final AHinter hinter = new ComplexFisherman(Tech.MutantSwordfish);
	getHints(MUTANT_SWORDFISH, hinter);
	AHint hint = find(HINTS, "Finned Mutant Swordfish: box 3, row 4, col C and row 1, row 9, col G on 2");
if ( REPRODUCE ) {
//KEEP THESE to maintain this testcase
System.out.format("hint        : %s\n", hint.toFullString());
System.out.format("clueHtml    : %s\n", hint.getClueHtml(true));
System.out.format("aquaCells   : %s\n", Frmt.csv(hint.getAquaCells(0)));
System.out.format("greens      : %s\n", hint.getGreens(0));
System.out.format("blue fin    : %s\n", hint.getBlues(null, 0));
System.out.format("oranges     : %s\n", hint.getOranges(0));
System.out.format("purple endo : %s\n", hint.getPurples());
System.out.format("yellow shark: %s\n", hint.getYellows());
System.out.format("reds        : %s\n", hint.getReds(0));
System.out.format("redPots     : %s\n", hint.redPots);
System.out.format("bases       : %s\n", Frmt.csv(hint.getBases()));
System.out.format("covers      : %s\n", Frmt.csv(hint.getCovers()));
}
	checkClue(hint, "Look for a Finned Mutant Swordfish on <b>2</b>");
	assertEquals(cellSet("C1, G1, I1, G3, G4, C9"), hint.getAquaCells(0)); // 4th corner is the fins
	assertEquals(newPots("C1-2, C9-2, G1-2, G3-2, G4-2, I1-2"), hint.getGreens(0));
	assertEquals(newPots("A4-2, C7-2"), hint.getBlues(null,0)); // the fins
	assertEquals(null, hint.getOranges(0));
	assertEquals(new Pots(), hint.getPurples()); // endo-fins: empty
	assertEquals(null, hint.getYellows()); // sharks
	assertEquals(newPots("A9-6"), hint.getReds(0));
	assertEquals(newPots("A9-6"), hint.redPots);
	assertArrayEquals(regions("box 3, row 4, col C"), hint.getBasesArray());
	assertArrayEquals(regions("row 1, row 9, col G"), hint.getCoversArray());
	assertHintEqualsStoredHintsHtml(this, hint, testName+".html", REPRODUCE);
}

/*
59#C:\Users\User\Documents\NetBeansProjects\DiufSudoku\top1465.d5.mt
1.2..9..386..319.2973652481.1.39....5.......7..8.27..9..19......8..7..95......3.4
,45,,478,48,,567,567,,,,45,47,,,,57,,,,,,,,,,,2467,,47,,,4568,258,245,68,,2349,469,148,1468,468,268,1234,,346,34,,15,,,156,134,,2347,2345,,,46,3456,2678,267,68,2346,,46,124,,346,126,,,267,259,5679,1258,168,568,,1267,
34   	    140,911,900	36	 135	  1	Mutant Jellyfish              	Finned Mutant Jellyfish: 6 r68c35 r59c7b8 fr6c1 fr8c1 efr8c3 => r9c1<>6 (A9-6)

138#C:\Users\User\Documents\NetBeansProjects\DiufSudoku\top1465.d5.mt
.....9.8.....4.....271..5..5.2.9.8.7.867.29.597185....2.89.56...5.62...8......25.
136,1346,345,25,367,,1347,,12346,1368,369,359,25,,378,137,236,1236,3468,,,,368,38,,3469,349,,34,,34,,16,,16,,34,,,,13,,,134,,,,,,,346,34,2346,2346,,134,,,137,,,347,134,137,,349,,,1347,134,3479,,167,13469,349,34,178,13478,,,1349
25   	    540,584,000	34	 146	  1	Mutant Jellyfish              	Finned Mutant Jellyfish: row 3, row 5, row 7, col C and box 7, row 1, col A, col H on 4 (I1-4)
*/
private static final String MUTANT_JELLYFISH
="1.2..9..386..319.2973652481.1.39....5.......7..8.27..9..19......8..7..95......3.4" + NL
+",45,,478,48,,567,567,,,,45,47,,,,57,,,,,,,,,,,2467,,47,,,4568,258,245,68,,2349,469,148,1468,468,268,1234,,346,34,,15,,,156,134,,2347,2345,,,46,3456,2678,267,68,2346,,46,124,,346,126,,,267,259,5679,1258,168,568,,1267,";
/** Test ComplexFisherman Tech.MutantJellyfish */
@Test
public void testMutantJellyfish() {
	final String testName = testName(this);
	final AHinter hinter = new ComplexFisherman(Tech.MutantJellyfish);
	getHints(MUTANT_JELLYFISH, hinter);
	AHint hint = find(HINTS, "Finned Mutant Jellyfish: row 6, row 8, col C, col E and box 8, row 5, row 9, col G on 6");
if ( REPRODUCE ) {
//KEEP THESE to maintain this testcase
System.out.format("hint        : %s\n", hint.toFullString());
System.out.format("aquaCells   : %s\n", Frmt.csv(hint.getAquaCells(0)));
System.out.format("oranges     : %s\n", hint.getOranges(0));
System.out.format("greens      : %s\n", hint.getGreens(0));
System.out.format("blue exoFin : %s\n", hint.getBlues(null, 0));
System.out.format("purple endo : %s\n", hint.getPurples());
System.out.format("yellow shark: %s\n", hint.getYellows());
System.out.format("reds        : %s\n", hint.getReds(0));
System.out.format("redPots     : %s\n", hint.redPots);
System.out.format("clueHtml    : %s\n", hint.getClueHtml(true));
System.out.format("bases       : %s\n", Frmt.csv(hint.getBases()));
System.out.format("covers      : %s\n", Frmt.csv(hint.getCovers()));
}
	assertEquals(null, hint.cell);
	assertEquals(0, hint.value);
	checkClue(hint, "Look for a Finned Mutant Jellyfish on <b>6</b>");
	assertEquals(cellSet("C5, E5, G6, E7, F8, G8, C9, E9"), hint.getAquaCells(0)); // 4th corner is the fins
	assertEquals(newPots("C5-6, C9-6, E5-6, E7-6, E9-6, F8-6, G6-6, G8-6"), hint.getGreens(0));
	assertEquals(newPots("A6-6, A8-6, C8-6"), hint.getBlues(null,0)); // the fins
	assertEquals(null, hint.getOranges(0));
	assertEquals(newPots("C8-6"), hint.getPurples()); // endo-fins: efr8c3
	assertEquals(null, hint.getYellows()); // sharks
	assertEquals(newPots("A9-6"), hint.getReds(0));
	assertEquals(newPots("A9-6"), hint.redPots);
	assertArrayEquals(regions("row 6, row 8, col C, col E"), hint.getBasesArray());
	assertArrayEquals(regions("box 8, row 5, row 9, col G"), hint.getCoversArray());
	assertHintEqualsStoredHintsHtml(this, hint, testName+".html", REPRODUCE);
}

}
