/*
 * Project: Sudoku Explainer
 * Copyright (C) 2006-2007 Nicolas Juillerat
 * Copyright (C) 2013-2020 Keith Corlett
 * Available under the terms of the Lesser General Public License (LGPL)
 */
package diuf.sudoku.solver.hinters.fish;

import static org.junit.Assert.*;
import static diuf.sudoku.TestHelp.*;
import org.junit.Test;
import diuf.sudoku.Tech;
import diuf.sudoku.Pots;
import diuf.sudoku.TestHelp;
import diuf.sudoku.solver.AHint;
import diuf.sudoku.solver.hinters.AHinter;

/**
 * Test the KrakenFisherman class, especially the findHints method.
 *
 * @author Keith Corlett 2020-10-09
 */
public class KrakenFishermanTest {

static {
	// so that TestHelp.getHints rebuilds all the s__t in the grid.
	TestHelp.REBUILD_GRID = true;
}

private static final boolean REPRODUCE = REPRODUCE_ALL_TESTS || false;

// ================================== SWAMPFISH ===============================

private static final String KRAKEN_SWAMPFISH
="53..1972..79..2.....25..9.....2..68.72...6.91.8..9.2.7.6172...9..7138.622..96.17." + NL
+",,468,468,,,,,468,1468,,,3468,48,,3458,135,4568,1468,14,,,478,37,,134,3468,1349,149,345,,457,137,,,345,,,345,348,458,,345,,,136,,456,34,,13,,345,,38,,,,,45,3458,345,,49,459,,,,,45,,,,45,38,,,45,,,38";
//459#C:\Users\User\Documents\NetBeansProjects\DiufSudoku\top1465.d5.mt
//NEW: Kraken type 2: Finned Swampfish: col D, col H and row 2, row 6 on 3 (G2-5)

/** Test {@link KrakenFisherman#findHints} seeking {@link Tech.KrakenSwampfish}. */
@Test
public void testKrakenSwampfish() {
	final String testName = testName(this);
	AHinter hinter = new KrakenFisherman(Tech.KrakenSwampfish);
	getHints(KRAKEN_SWAMPFISH, hinter);
	AHint hint = find(HINTS, "Kraken type 2: Finned Swampfish: col D, col H and row 2, row 6 on 3");
if ( REPRODUCE ) {
//KEEP THESE to maintain this testcase.
System.out.format("hint        : %s\n", hint.toFullString());
System.out.format("clueHtml    : %s\n", hint.getClueHtml(true));
System.out.format("aquaCells   : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getAquaCells(0)));
System.out.format("greens      : %s\n", hint.getGreens(0));
System.out.format("oranges     : %s\n", hint.getOranges(0));
System.out.format("blue exoFin : %s\n", hint.getBlues(null, 0));
System.out.format("purple endo : %s\n", hint.getPurples());
System.out.format("yellow shark: %s\n", hint.getYellows());
System.out.format("reds        : %s\n", hint.getReds(0));
System.out.format("bases       : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getBases()));
System.out.format("covers      : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getCovers()));
}
	assertEquals(null, hint.cell);
	assertEquals(0, hint.value);
	checkClue(hint, "Look for a Kraken type 2: Finned Swampfish in <b1>col D</b1>");
	assertEquals(cellSet("D2, H2, D6, H6"), hint.getAquaCells(0));
	assertEquals(newPots("D2-3, D6-3, H2-3, H6-3"), hint.getGreens(0));
	assertEquals(null, hint.getOranges(0));
	assertEquals(newPots("D5-3, H3-3, H7-3"), hint.getBlues(null,0)); // exo-fins
	assertEquals(Pots.EMPTY, hint.getPurples()); // endo-fins
	assertEquals(null, hint.getYellows()); // sharks
	assertEquals(newPots("G2-5"), hint.getReds(0));
	assertArrayEquals(regions("col D, col H"), hint.getBasesArray());
	assertArrayEquals(regions("row 2, row 6"), hint.getCoversArray());
	assertHintEqualsStoredHintsHtml(this, hint, testName+".html", REPRODUCE);
}


// ================================== SWORDFISH ===============================

private static final String KRAKEN_SWORDFISH
="4.123..6......7..16.81549.....32...4.....5.9.8...71..6..45.....91....6....3....2." + NL
+",579,,,,89,578,,578,235,2359,259,689,689,,23458,348,,,237,,,,,,37,237,157,59,5679,,,689,1578,1578,,1237,234,267,468,468,,12378,,2378,,234,259,49,,,235,35,,27,68,,,1689,23,1378,178,3789,,,257,478,48,23,,34578,3578,57,68,,46789,1469,689,1457,,5789";
//24#C:\Users\User\Documents\NetBeansProjects\DiufSudoku\top1465.d5.mt
//OLD: Kraken Fish Type 2: => r9c5<>6 Finned Franken Swordfish: 8 r2b69 c789 fr2c4 fr2c5 r2c4 =6= r2c5 -6- r9c5 r2c5 -8- r8c5 -4- r8c8 =4= r9c7 =1= r9c5 r5c9 -8- r4c78 =8= r4c6 =6= r9c6 -6- r9c5 r7c9 =9= r7c5 =1= r9c5 r8c9 -8- r8c5 -4- r8c8 =4= r9c7 =1= r9c5 r9c9 -8- r9c2 -6- r9c5 (E9-6)
//NEW: Kraken type 2: Finned Franken Swordfish: row 2, row 4, row 7 and box 2, col G, col H on 8 (E9-6)

/** Test {@link KrakenFisherman#findHints} seeking {@link Tech.KrakenSwordfish}. */
@Test
public void testKrakenSwordfish() {
	final String testName = testName(this);
	AHinter hinter = new KrakenFisherman(Tech.KrakenSwordfish);
	getHints(KRAKEN_SWORDFISH, hinter);
	AHint hint = find(HINTS, "Kraken type 2: Finned Franken Swordfish: row 2, row 4, row 7 and box 2, col G, col H on 8");
if ( REPRODUCE ) {
//KEEP THESE to maintain this testcase.
System.out.format("hint        : %s\n", hint.toFullString());
System.out.format("clueHtml    : %s\n", hint.getClueHtml(true));
System.out.format("aquaCells   : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getAquaCells(0)));
System.out.format("greens      : %s\n", hint.getGreens(0));
System.out.format("oranges     : %s\n", hint.getOranges(0));
System.out.format("blue exoFin : %s\n", hint.getBlues(null, 0));
System.out.format("purple endo : %s\n", hint.getPurples());
System.out.format("yellow shark: %s\n", hint.getYellows());
System.out.format("reds        : %s\n", hint.getReds(0));
System.out.format("bases       : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getBases()));
System.out.format("covers      : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getCovers()));
}
	assertEquals(null, hint.cell);
	assertEquals(0, hint.value);
	checkClue(hint, "Look for a Kraken type 2: Finned Franken Swordfish in <b1>row 2</b1>");
	assertEquals(cellSet("D2, E2, G2, H2, G4, H4, G7, H7"), hint.getAquaCells(0));
	assertEquals(newPots("D2-8, E2-8, G2-8, G4-8, G7-8, H2-8, H4-8, H7-8"), hint.getGreens(0));
	assertEquals(null, hint.getOranges(0));
	assertEquals(newPots("B7-8, E7-8, F4-8, I7-8"), hint.getBlues(null,0)); // exo-fins
	assertEquals(Pots.EMPTY, hint.getPurples()); // endo-fins
	assertEquals(null, hint.getYellows()); // sharks
	assertEquals(newPots("E9-6"), hint.getReds(0));
	assertArrayEquals(regions("row 2, row 4, row 7"), hint.getBasesArray());
	assertArrayEquals(regions("box 2, col G, col H"), hint.getCoversArray());
	assertHintEqualsStoredHintsHtml(this, hint, testName+".html", REPRODUCE);
}


// ================================== JELLYFISH ===============================

private static final String KRAKEN_JELLYFISH
="4.123..6......7..16.81549.....32...4.....5.9.8...71..6..45.....91....6....3....2." + NL
+",579,,,,89,578,,578,235,2359,259,689,689,,23458,348,,,237,,,,,,37,237,157,59,5679,,,689,1578,1578,,1237,234,267,468,468,,12378,,2378,,234,259,49,,,235,35,,27,68,,,1689,23,1378,178,3789,,,257,478,48,23,,34578,3578,57,68,,46789,1469,689,1457,,5789";
//459#C:\Users\User\Documents\NetBeansProjects\DiufSudoku\top1465.d5.mt
//OLD: Kraken Fish Type 2: => r2c7<>5 Finned Mutant Jellyfish: 3 c4b369 r6c789 fr2c4 fr5c4 r2c4 =6= r1c4 -6- r1c3 =6= r6c3 =5= r6c8 -5- r7c8 =5= r78c7 -5- r2c7 r2c8 =1= r2c1 -1- r3c2 -4- r9c2 -5- r8c2 =5= r8c7 -5- r2c7 r3c8 -3- r3c6 =3= r2c4 =6= r1c4 -6- r1c3 =6= r6c3 =5= r6c8 -5- r7c8 =5= r78c7 -5- r2c7 r5c4 =8= r5c5 =5= r4c5 -5- r4c9 =5= r2c9 -5- r2c7 r7c8 =5= r78c7 -5- r2c7 (G2-5)
//NEW: Kraken type 2: Finned Franken Jellyfish: box 7, box 9, row 2, row 4 and col B, col G, col H, col I on 8 (E9-6)

/** Test {@link KrakenFisherman#findHints} seeking {@link Tech.KrakenJellyfish}. */
@Test
public void testKrakenJellyfish() {
	final String testName = testName(this);
	AHinter hinter = new KrakenFisherman(Tech.KrakenJellyfish);
	getHints(KRAKEN_JELLYFISH, hinter);
	AHint hint = find(HINTS, "Kraken type 2: Finned Franken Jellyfish: box 7, box 9, row 2, row 4 and col B, col G, col H, col I on 8");
if ( REPRODUCE ) {
//KEEP THESE to maintain this testcase.
System.out.format("hint        : %s\n", hint.toFullString());
System.out.format("clueHtml    : %s\n", hint.getClueHtml(true));
System.out.format("aquaCells   : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getAquaCells(0)));
System.out.format("greens      : %s\n", hint.getGreens(0));
System.out.format("oranges     : %s\n", hint.getOranges(0));
System.out.format("blue exoFin : %s\n", hint.getBlues(null, 0));
System.out.format("purple endo : %s\n", hint.getPurples());
System.out.format("yellow shark: %s\n", hint.getYellows());
System.out.format("reds        : %s\n", hint.getReds(0));
System.out.format("bases       : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getBases()));
System.out.format("covers      : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getCovers()));
}
	assertEquals(null, hint.cell);
	assertEquals(0, hint.value);
	checkClue(hint, "Look for a Kraken type 2: Finned Franken Jellyfish in <b1>box 7</b1>");
	assertEquals(cellSet("G2, H2, G4, H4, B7, G7, H7, I7, H8, I8, B9, I9"), hint.getAquaCells(0));
	assertEquals(newPots("B7-8, B9-8, G2-8, G4-8, G7-8, H2-8, H4-8, H7-8, H8-8, I7-8, I8-8, I9-8"), hint.getGreens(0));
	assertEquals(null, hint.getOranges(0));
	assertEquals(newPots("D2-8, E2-8, F4-8"), hint.getBlues(null,0)); // exo-fins
	assertEquals(Pots.EMPTY, hint.getPurples()); // endo-fins
	assertEquals(null, hint.getYellows()); // sharks
	assertEquals(newPots("E9-6"), hint.getReds(0));
	assertArrayEquals(regions("box 7, box 9, row 2, row 4"), hint.getBasesArray());
	assertArrayEquals(regions("col B, col G, col H, col I"), hint.getCoversArray());
	assertHintEqualsStoredHintsHtml(this, hint, testName+".html", REPRODUCE);
}

}
