/*
 * Project: Sudoku Explainer
 * Copyright (C) 2006-2007 Nicolas Juillerat
 * Copyright (C) 2013-2020 Keith Corlett
 * Available under the terms of the Lesser General Public License (LGPL)
 */
package diuf.sudoku.solver.hinters.nkdset;

import static org.junit.Assert.*;
import static diuf.sudoku.TestHelp.*;
import org.junit.Test;
import diuf.sudoku.Pots;
import diuf.sudoku.Tech;
import diuf.sudoku.Values;
import diuf.sudoku.solver.AHint;
import diuf.sudoku.solver.hinters.AHinter;
import diuf.sudoku.utils.Frmt;

/**
 * Tests the NakedSet hinter.
 * @author Keith Corlett 2018 Apr
 */
public final class NakedSetTest {
	
private static final boolean REPRODUCE = REPRODUCE_ALL_TESTS || false;

private static final String NAKED_PAIR
="7.....4...2..7..8...3..8..9...5.63...6..2..9...1..7..6...3.29...3..4..6...9..1.35" + NL
+",89,68,26,13,359,,25,13,1569,,56,149,,349,56,,13,14,145,,26,15,,267,27,,289,789,247,,189,,,14,2478,358,,457,18,,34,1578,,478,2359,589,,48,39,,258,245,,1568,1578,5678,,568,,,147,478,158,,2578,789,,59,178,,1278,2468,48,,78,68,,278,,";
//1#C:\Users\User\Documents\NetBeansProjects\DiufSudoku\top1465.d3.mt

/**
 * Naked Pair
 */
@Test
public void testNakedPair() {
	final String testName = testName(this);
	AHinter hinter = new NakedSet(Tech.NakedPair);
	getHints(NAKED_PAIR, hinter);

	// HINT 1
	AHint hint = find(HINTS, "Naked Pair: E1, I1: 1, 3 in row 1", "NakedSetHint");
if ( REPRODUCE ) {
//KEEP THESE to maintain this testcase.
System.out.format("aquaCells  : %s\n", Frmt.csv(hint.getAquaCells(0)));
System.out.format("oranges    : %s\n", hint.getOranges(0));
System.out.format("greens     : %s\n", hint.getGreens(0));
System.out.format("reds       : %s\n", hint.getReds(0));
System.out.format("redPots    : %s\n", hint.redPots);
System.out.format("clueHtml   : %s\n", hint.getClueHtml(true));
System.out.format("bases      : %s\n", Frmt.csv(hint.getBases()));
System.out.format("covers     : %s\n", Frmt.csv(hint.getCovers()));
}
	assertEquals(grid.getCSVSet("E1, I1"), hint.getAquaCells(0));
	assertEquals(newPots("E1-13, I1-13"), hint.getGreens(0));
	assertEquals(newPots("F1-3"), hint.getReds(0));
	assertEquals(newPots("F1-3"), hint.redPots);
	assertEquals(grid.rows[0], hint.getBases().get(0));
	assertEquals(null, hint.getCovers());
	checkClue(hint, "Look for a Naked Pair in <b1>row 1</b1>");
	assertHintEqualsStoredHintsHtml(this, hint, testName+"_1.html", REPRODUCE, false);

	// HINT 2
	hint = find(HINTS , "Naked Pair: C2, G2: 5, 6 in row 2");
if ( REPRODUCE ) {
//KEEP THESE to maintain this testcase.
System.out.format("aquaCells  : %s\n", Frmt.csv(hint.getAquaCells(0)));
System.out.format("oranges    : %s\n", hint.getOranges(0));
System.out.format("greens     : %s\n", hint.getGreens(0));
System.out.format("reds       : %s\n", hint.getReds(0));
System.out.format("redPots    : %s\n", hint.redPots);
System.out.format("clueHtml   : %s\n", hint.getClueHtml(true));
System.out.format("bases      : %s\n", Frmt.csv(hint.getBases()));
System.out.format("covers     : %s\n", Frmt.csv(hint.getCovers()));
}
	assertEquals(grid.getCSVSet("C2, G2"), hint.getAquaCells(0));
	assertEquals(newPots("C2-56 G2-56"), hint.getGreens(0));
	assertEquals(new Pots(grid.get("A2"), new Values(5,6)), hint.getReds(0));
	assertEquals(new Pots(grid.get("A2"), new Values(5,6)), hint.redPots);
	assertEquals(grid.rows[1], hint.getBases().get(0));
	assertEquals(null, hint.getCovers());
	checkClue(hint, "Look for a Naked Pair in <b1>row 2</b1>");
	assertHintEqualsStoredHintsHtml(this, hint, testName+"_2.html", REPRODUCE, false);

	// HINT 3
	hint = find(HINTS , "Naked Pair: I1, I2: 1, 3 in col I");
if ( REPRODUCE ) {
//KEEP THESE to maintain this testcase.
System.out.format("clueHtml   : %s\n", hint.getClueHtml(true));
System.out.format("aquaCells  : %s\n", Frmt.csv(hint.getAquaCells(0)));
System.out.format("oranges    : %s\n", hint.getOranges(0));
System.out.format("greens     : %s\n", hint.getGreens(0));
System.out.format("reds       : %s\n", hint.getReds(0));
System.out.format("redPots    : %s\n", hint.redPots);
System.out.format("bases      : %s\n", Frmt.csv(hint.getBases()));
System.out.format("covers     : %s\n", Frmt.csv(hint.getCovers()));
}
	checkClue(hint, "Look for a Naked Pair in <b1>col I</b1>");
	assertEquals(grid.getCSVSet("I1, I2"), hint.getAquaCells(0));
	assertEquals(null, hint.getOranges(0));
	assertEquals(newPots("I1-13, I2-13"), hint.getGreens(0));
	assertEquals(newPots("I8-1"), hint.getReds(0));
	assertEquals(newPots("I8-1"), hint.redPots);
	assertEquals(grid.cols[8], hint.getBases().get(0));
	assertEquals(null, hint.getCovers());
	assertHintEqualsStoredHintsHtml(this, hint, testName+"_3.html", REPRODUCE); // fails if PRODUCE is true
}


private static final String DIRECT_NAKED_PAIR
="7.8...3..93.651...5...3.....4..9..263.9.8.......1...93.9.2....4.8..7.5..........." + NL
+",16,,49,24,249,,156,15,,,24,,,,247,478,78,,126,1246,78,,78,2469,146,129,18,,17,35,,35,178,,,,26,,47,,26,147,1457,157,268,57,2567,,246,2467,478,,,16,,1357,,16,3568,167,13678,,1246,,1236,349,,3469,,136,129,1246,57,123567,34589,146,345689,269,13678,12789";
//4#C:\Users\User\Documents\NetBeansProjects\DiufSudoku\top1465.d3.mt

/**
 * Direct Naked Pair
 */
@Test
public void testDirectNakedPair() {
	final String testName = testName(this);
	AHinter hinter = new NakedSet(Tech.DirectNakedPair); // DIFFERENT TECH!
	getHints(DIRECT_NAKED_PAIR, hinter);
	AHint hint = find(HINTS, "Direct Naked Pair: A7, E7: 1, 6 in row 7", "NakedSetDirectHint");
if ( REPRODUCE ) {
//KEEP THESE to maintain this testcase.
System.out.format("clueHtml   : %s\n", hint.getClueHtml(true));
System.out.format("aquaCells  : %s\n", Frmt.csv(hint.getAquaCells(0)));
System.out.format("oranges    : %s\n", hint.getOranges(0));
System.out.format("greens     : %s\n", hint.getGreens(0));
System.out.format("reds       : %s\n", hint.getReds(0));
System.out.format("redPots    : %s\n", hint.redPots);
System.out.format("bases      : %s\n", Frmt.csv(hint.getBases()));
System.out.format("covers     : %s\n", Frmt.csv(hint.getCovers()));
}
	assertEquals(true, hint.getClass().getSimpleName().contains("Direct"));
	assertEquals(grid.get("G7"), hint.cell);
	assertEquals(7, hint.value);
	checkClue(hint, "Look for a Direct Naked Pair in <b1>row 7</b1>");
	assertEquals(grid.getCSVSet("G7"), hint.getAquaCells(0));
	assertEquals(newPots("A7-16, E7-16"), hint.getOranges(0));
	assertEquals(newPots("G7-7"), hint.getGreens(0));
	assertEquals(newPots("C7-1, F7-6, G7-16, H7-16"), hint.getReds(0));
	assertEquals(newPots("C7-1, F7-6, G7-16, H7-16"), hint.redPots);
	assertEquals(grid.rows[6], hint.getBases().get(0));
	assertEquals(null, hint.getCovers());
	assertHintEqualsStoredHintsHtml(this, hint, testName+".html", REPRODUCE);
}


private static final String NAKED_TRIPLE
="7.....4...2..7..8.4.3..8..9.7.5.63...6..2..9...1..7..6...3.29...3..4..6..49..1.35" + NL
+",89,68,26,13,59,,25,13,19,,56,149,,34,56,,13,,15,,26,15,,267,27,,89,,24,,189,,,14,248,35,,45,18,,34,1578,,78,235,89,,48,39,,258,245,,1568,15,78,,568,,,147,478,158,,2578,789,,59,178,,278,268,,,78,68,,278,,";
//1#C:\Users\User\Documents\NetBeansProjects\DiufSudoku\top1465.d3.mt

/**
 * Naked Triple
 */
@Test
public void testNakedTriple() {
	final String testName = testName(this);
	AHinter hinter = new NakedSet(Tech.NakedTriple);
	getHints(NAKED_TRIPLE, hinter);
	AHint hint = find(HINTS, "Naked Triple: A5, C5, F5: 3, 4, 5 in row 5", "NakedSetHint");
if ( REPRODUCE ) {
//KEEP THESE to maintain this testcase.
System.out.format("clueHtml   : %s\n", hint.getClueHtml(true));
System.out.format("aquaCells  : %s\n", Frmt.csv(hint.getAquaCells(0)));
System.out.format("oranges    : %s\n", hint.getOranges(0));
System.out.format("greens     : %s\n", hint.getGreens(0));
System.out.format("reds       : %s\n", hint.getReds(0));
System.out.format("redPots    : %s\n", hint.redPots);
System.out.format("bases      : %s\n", Frmt.csv(hint.getBases()));
System.out.format("covers     : %s\n", Frmt.csv(hint.getCovers()));
}
	checkClue(hint, "Look for a Naked Triple in <b1>row 5</b1>");
	assertEquals(grid.getCSVSet("A5, C5, F5"), hint.getAquaCells(0));
	assertEquals(newPots("A5-35 C5-45 F5-34"), hint.getGreens(0));
	assertEquals(new Pots(grid.get("G5"), new Values(5)), hint.getReds(0));
	assertEquals(new Pots(grid.get("G5"), new Values(5)), hint.redPots);
	assertEquals(grid.rows[4], hint.getBases().get(0));
	assertEquals(null, hint.getCovers());
	assertHintEqualsStoredHintsHtml(this, hint, testName+".html", REPRODUCE);
}


private static final String NAKED_QUAD
="1..5.36.283269.75..5.72...37.9..2....1....9...8.9..........63..3.1....675..37..1." + NL
+",79,47,,48,,,89,,,,,,,14,,,14,69,,46,,,18,18,49,,,46,,148,1368,,15,34,568,246,,35,48,3468,57,,2347,68,246,,35,,1346,57,124,2347,16,49,27,78,124,145,,,248,59,,49,,248,458,89,245,,,,26,68,,,49,28,,49";
//5#C:\Users\User\Documents\NetBeansProjects\DiufSudoku\top1465.d3.mt

/**
 * Naked Quad
 */
@Test
public void testNakedQuad() {
	final String testName = testName(this);
	AHinter hinter = new NakedSet(Tech.NakedQuad);
	getHints(NAKED_QUAD, hinter);
	AHint hint = find(HINTS, "Naked Quad: G4, I4, I5, I6: 1, 5, 6, 8 in box 6", "NakedSetHint");
if ( REPRODUCE ) {
//KEEP THESE to maintain this testcase.
System.out.format("clueHtml   : %s\n", hint.getClueHtml(true));
System.out.format("aquaCells  : %s\n", Frmt.csv(hint.getAquaCells(0)));
System.out.format("oranges    : %s\n", hint.getOranges(0));
System.out.format("greens     : %s\n", hint.getGreens(0));
System.out.format("reds       : %s\n", hint.getReds(0));
System.out.format("redPots    : %s\n", hint.redPots);
System.out.format("bases      : %s\n", Frmt.csv(hint.getBases()));
System.out.format("covers     : %s\n", Frmt.csv(hint.getCovers()));
}
	checkClue(hint, "Look for a Naked Quad in <b1>box 6</b1>");
	assertEquals(grid.getCSVSet("G4, I4, I5, I6"), hint.getAquaCells(0));
	assertEquals(newPots("G4-15 I4-568 I5-68 I6-16"), hint.getGreens(0));
	assertEquals(new Pots(grid.get("G6"), new Values(1)), hint.getReds(0));
	assertEquals(new Pots(grid.get("G6"), new Values(1)), hint.redPots);
	assertEquals(grid.boxs[5], hint.getBases().get(0));
	assertEquals(null, hint.getCovers());
	assertHintEqualsStoredHintsHtml(this, hint, testName+".html", REPRODUCE);
}

}
