/*
 * Project: Sudoku Explainer
 * Copyright (C) 2006-2007 Nicolas Juillerat
 * Copyright (C) 2013-2020 Keith Corlett
 * Available under the terms of the Lesser General Public License (LGPL)
 */
package diuf.sudoku.solver.hinters.urt;

import static org.junit.Assert.*;
import static diuf.sudoku.TestHelp.*;
import org.junit.Test;
import diuf.sudoku.Pots;
import diuf.sudoku.Values;
import diuf.sudoku.solver.AHint;
import diuf.sudoku.solver.hinters.AHinter;
import diuf.sudoku.utils.MyLinkedHashSet;


/**
 * Test the UniqueRectangles hinter.
 * @author Keith Corlett 2018 Apr
 */
public final class UniqueRectanglesTest {

private static final boolean REPRODUCE = REPRODUCE_ALL_TESTS || false;

private static final String URT1
="..3..9.8.8..3....2.675...3...27..64..3..45.1.....1...5..1....9...5..71..49......." + NL
+"15,1245,,1246,267,,457,,1467,,145,49,,67,146,4579,567,,29,,,,28,1248,49,,149,15,15,,,389,38,,,389,679,,689,2689,,,2789,,789,679,478,4689,2689,,2368,23789,27,,2367,278,,2468,23568,23468,23458,,3468,23,28,,24689,2369,,,26,3468,,,68,1268,23568,12368,23578,257,378";
//16#C:\Users\User\Documents\NetBeansProjects\DiufSudoku\top1465.d3.mt
// Unique Rectangle type 1: A1, B1, B4, A4 on 1, 5 (B1-15)

private AHinter hinter = new UniqueRectangle();

/** Test Unique Rectangle type 1. */
@Test
public void testURT1() {
	String testName = testName(this);
	getHints(URT1, hinter);
	AHint hint = find(HINTS, "Unique Rectangle type 1: A1, B1, B4, A4 on 1 and 5");
////KEEP THESE: For use in creating test cases.
//System.out.format("aquaCells  : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getAquaCells(0)));
//System.out.format("oranges    : %s\n", hint.getOranges(0));
//System.out.format("greens     : %s\n", hint.getGreens(0));
//System.out.format("reds       : %s\n", hint.getReds(0));
//System.out.format("redPots    : %s\n", hint.redPots);
//System.out.format("clueHtml   : %s\n", hint.getClueHtml(true));
//System.out.format("bases      : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getBases()));
//System.out.format("covers     : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getCovers()));
	checkClue(hint, "Look for a Unique Rectangle type 1 on <b>1</b> and <b>5</b>");
	assertEquals(grid.getCSVSet("A1, A4, B1, B4"), hint.getAquaCells(0));
	assertEquals(newPots("A1-15, A4-15, B4-15"), hint.getGreens(0));
	assertEquals(new Pots(grid.get("B1"), new Values(1,5)), hint.getReds(0));
	assertEquals(new Pots(grid.get("B1"), new Values(1,5)), hint.redPots);
	assertEquals(link("A1","B1","B4","A4") , hint.getLinks(0));
	assertEquals(null, hint.getBases());
	assertEquals(null, hint.getCovers());
	assertHintEqualsStoredHintsHtml(this, hint, testName+".html", REPRODUCE);
}


private static final String URT2
="9...182....5267....7.93.....29..3.8.3..6..7.27...2........9....1.73.2..6..8..6..5" + NL
+",346,36,45,,,,4567,47,48,14,,,,,13489,1349,3489,2468,,1246,,,45,14568,1456,148,56,,,147,47,,56,,14,,1458,14,,58,1459,,1459,,,1468,146,1458,,1459,134569,134569,1349,2456,3456,2346,14578,,145,1348,1237,3478,,459,,,58,,489,49,,24,349,,147,47,,1349,123479,";
//20#C:\Users\User\Documents\NetBeansProjects\DiufSudoku\top1465.d3.mt
// Unique Rectangle type 2: E4, D4, D9, E9 on 4 and 7 (D6-1, D7-1)

/** Test Unique Rectangle type 2. */
@Test
public void testURT2() {
	final String testName = testName(this);
	getHints(URT2, hinter);
	AHint hint = find(HINTS, "Unique Rectangle type 2: E4, D4, D9, E9 on 4 and 7");
////KEEP THESE: For use in creating test cases.
//System.out.format("aquaCells  : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getAquaCells(0)));
//System.out.format("oranges    : %s\n", hint.getOranges(0));
//System.out.format("greens     : %s\n", hint.getGreens(0));
//System.out.format("reds       : %s\n", hint.getReds(0));
//System.out.format("redPots    : %s\n", hint.redPots);
//System.out.format("clueHtml   : %s\n", hint.getClueHtml(true));
//System.out.format("bases      : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getBases()));
//System.out.format("covers     : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getCovers()));
	checkClue(hint, "Look for a Unique Rectangle type 2 on <b>4</b> and <b>7</b>");
	assertEquals(grid.getCSVSet("E4, D4, D9, E9"), hint.getAquaCells(0));
	assertEquals(newPots("D4-1, D9-1"), hint.getOranges(0));
	assertEquals(newPots("D4-47, D9-47, E4-47, E9-47"), hint.getGreens(0));
	assertEquals(newPots("D6-1, D7-1"), hint.getReds(0));
	assertEquals(newPots("D6-1, D7-1"), hint.redPots);
	assertEquals(link("E4","D4","D9","E9"), hint.getLinks(0));
	assertEquals(null, hint.getBases());
	assertEquals(null, hint.getCovers());
	assertHintEqualsStoredHintsHtml(this, hint, testName+".html", REPRODUCE);
}


private static final String URT3
="...6...94.753.9.619...8.735519.364.86.3..8...78.91.3.625...3.....7...58..9....6.3" + NL
+"13,23,128,,57,57,28,,,48,,,,24,,28,,,,246,46,124,,12,,,,,,,27,,,,27,,,24,,2457,2457,,19,1257,279,,,24,,,245,,25,,,,1468,148,69,,19,147,79,134,346,,14,69,124,,,29,148,,148,124578,2457,2457,,124,";
//8#C:\Users\User\Documents\NetBeansProjects\DiufSudoku\top1465.d3.mt
// Unique Rectangle type 3: E1, F1, F9, E9 on 5, 7 (D9-1248)

/** Test Unique Rectangle type 3. */
@Test
public void testURT3() {
	final String testName = testName(this);
	getHints(URT3, hinter);
	// HINT 1
	AHint hint = find(HINTS, "Unique Rectangle type 3 Hidden Set: E1, F1, F9, E9 on 5 and 7");
////KEEP THESE: For use in creating test cases.
//System.out.format("aquaCells  : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getAquaCells(0)));
//System.out.format("oranges    : %s\n", hint.getOranges(0));
//System.out.format("greens     : %s\n", hint.getGreens(0));
//System.out.format("reds       : %s\n", hint.getReds(0));
//System.out.format("redPots    : %s\n", hint.redPots);
//System.out.format("clueHtml   : %s\n", hint.getClueHtml(true));
//System.out.format("bases      : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getBases()));
//System.out.format("covers     : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getCovers()));
	checkClue(hint, "Look for a Unique Rectangle type 3 Hidden Set on <b>5</b> and <b>7</b>");
	assertEquals(grid.getCSVSet("E1, F1, F9, E9"), hint.getAquaCells(0));
	assertEquals(newPots("D9-57, E9-57, F9-57"), hint.getOranges(0));
	assertEquals(newPots("E1-57, E9-57, F1-57, F9-57"), hint.getGreens(0));
	assertEquals(newPots("D9-1248"), hint.getReds(0));
	assertEquals(newPots("D9-1248"), hint.redPots);
	assertEquals(link("E1","F1","F9","E9"), hint.getLinks(0));
	assertEquals(grid.boxs[7], hint.getBases().get(0));
	assertEquals(null, hint.getCovers());
	assertHintEqualsStoredHintsHtml(this, hint, testName+"_1.html", REPRODUCE, false);
	// CHECK SUBSEQUENT HINTS (the quick and dirty way)
	// the order changed, so I've replaced HashSet with LinkedHashSet because
	// somebody (me) changed how cell hashCode is calculated without thinking
	// through all the ramifications. A LinkedHashMap will be consistent.
	for ( int i=2; i<=5; ++i )
		assertHintEqualsStoredHintsHtml(this, accu.getHint(), testName+"_"+i+".html", REPRODUCE, false);

}


private static final String URT4
=".....8.7..134.6.5..8...1..4...1.45...9...73....82..7.65..6.29...2..43.65.6..1.4.." + NL
+"269,45,24569,359,259,,126,,139,279,,,,279,,28,,89,2679,,25679,379,2579,,26,39,,2367,37,267,,368,,,289,289,1246,,1256,58,568,,,1248,128,14,35,,,39,59,,14,,,347,14,,78,,,138,1378,1789,,179,789,,,18,,,3789,,79,5789,,59,,238,2378";
//12#C:\Users\User\Documents\NetBeansProjects\DiufSudoku\top1465.d3.mt
// Unique Rectangle type 4: A6, A5, H5, H6 on 4, 1 (A5-1, H5-1)

/** Test Unique Rectangle type 4. */
@Test
public void testURT4() {
	final String testName = testName(this);
	getHints(URT4, hinter);
	AHint hint = find(HINTS, "Unique Rectangle type 4: A6, A5, H5, H6 on 4 and 1"); // (A5-1, H5-1)
////KEEP THESE: For use in creating test cases.
//System.out.format("aquaCells  : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getAquaCells(0)));
//System.out.format("oranges    : %s\n", hint.getOranges(0));
//System.out.format("greens     : %s\n", hint.getGreens(0));
//System.out.format("reds       : %s\n", hint.getReds(0));
//System.out.format("redPots    : %s\n", hint.redPots);
//System.out.format("clueHtml   : %s\n", hint.getClueHtml(true));
//System.out.format("bases      : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getBases()));
//System.out.format("covers     : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getCovers()));
	checkClue(hint, "Look for a Unique Rectangle type 4 on <b>4</b> and <b>1</b>");
	assertEquals(grid.getCSVSet("A6, A5, H5, H6"), hint.getAquaCells(0));
	assertEquals(newPots("A5-4, A6-14, H5-4, H6-14"), hint.getGreens(0));
	assertEquals(newPots(1, "A5","H5"), hint.getReds(0));
	assertEquals(newPots(1, "A5","H5"), hint.redPots);
	assertEquals(link("A6","A5","H5","H6"), hint.getLinks(0));
	assertEquals(grid.rows[4], hint.getBases().get(0));
	assertEquals(null, hint.getCovers());
	assertHintEqualsStoredHintsHtml(this, hint, testName+".html", REPRODUCE);
}


private static final String LOOP_TYPE_4
="91......44.5...9..2..984135.2164..79.4.3...1.5.9271.48....2.4..1.48.3.6...24.78.." + NL
+",,38,57,36,256,67,28,,,38,,17,136,26,,28,67,,67,67,,,,,,,38,,,,,58,35,,,678,,678,,59,589,256,,26,,36,,,,,36,,,3678,3678,3678,15,,569,,59,13,,59,,,59,,27,,27,36,59,,,16,,,59,13";
//C:\Users\User\Documents\SodukuPuzzles\MagicTour\top1465.d3.0199.UniqueLoop8.mt.txt
// Unique Loop 8 type 4: E5, F5, F7, H7, H9, B9, B8, E8 on 9 and 5

/** Test Unique Loop type 4. */
@Test
public void testLoop4() {
	final String testName = testName(this);
	getHints(LOOP_TYPE_4, hinter);
	AHint hint = find(HINTS, "Unique Loop 8 type 4: E5, F5, F7, H7, H9, B9, B8, E8 on 9 and 5");
////KEEP THESE: For use in creating test cases.
//System.out.format("aquaCells  : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getAquaCells(0)));
//System.out.format("oranges    : %s\n", hint.getOranges(0));
//System.out.format("greens     : %s\n", hint.getGreens(0));
//System.out.format("reds       : %s\n", hint.getReds(0));
//System.out.format("redPots    : %s\n", hint.redPots);
//System.out.format("clueHtml   : %s\n", hint.getClueHtml(true));
//System.out.format("bases      : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getBases()));
//System.out.format("covers     : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getCovers()));
	checkClue(hint, "Look for a Unique Loop 8 type 4 on <b>9</b> and <b>5</b>");
	assertEquals(grid.getCSVSet("E5, F5, F7, H7, H9, B9, B8, E8"), hint.getAquaCells(0));
	assertEquals(newPots("B8-59, B9-59, E5-59, E8-59, F5-9, F7-9, H7-59, H9-59"), hint.getGreens(0));
	assertEquals(newPots(5, "F5","F7"), hint.getReds(0));
	assertEquals(newPots(5, "F5","F7"), hint.redPots);
	assertEquals(link("E5", "F5", "F7", "H7", "H9", "B9", "B8", "E8"), hint.getLinks(0));
	assertEquals(grid.cols[5], hint.getBases().get(0));
	assertEquals(null, hint.getCovers());
	assertHintEqualsStoredHintsHtml(this, hint, testName+".html", REPRODUCE);
}

}
