/*
 * Project: Sudoku Explainer
 * Copyright (C) 2006-2007 Nicolas Juillerat
 * Copyright (C) 2013-2020 Keith Corlett
 * Available under the terms of the Lesser General Public License (LGPL)
 */
package diuf.sudoku.solver.hinters.wing;

import static org.junit.Assert.*;
import static diuf.sudoku.TestHelp.*;
import diuf.sudoku.Grid.Cell;
import diuf.sudoku.Link;
import diuf.sudoku.Tech;
import diuf.sudoku.solver.AHint;
import diuf.sudoku.solver.hinters.AHinter;
import diuf.sudoku.utils.MyLinkedHashSet;
import org.junit.Test;
import java.util.List;
import java.util.ArrayList;

/**
 * Test the XYWing hinter.
 * @author Keith Corlett 2018 Apr
 */
public final class XYWingTest {
	
private static final boolean REPRODUCE = REPRODUCE_ALL_TESTS;

private static final String XY_WING
="7.....4...2..7..8.4.3..8..9.7.5.631..6.12.897..1..7..6...3.29...3..4.16..49..1.35" + NL
+",89,68,26,13,59,,25,13,19,,56,49,,34,56,,13,,15,,26,15,,267,27,,89,,24,,89,,,,24,35,,45,,,34,,,,23,89,,48,39,,25,245,,16,15,78,,56,,,47,48,58,,2578,789,,59,,,28,268,,,78,68,,27,,";
//1#C:\Users\User\Documents\NetBeansProjects\DiufSudoku\top1465.d3.mt

/** Test XY-Wing */
@Test
public void testXYWing() {
	final String testName = testName(this);
	AHinter hinter = new XYWing(Tech.XY_Wing);
	getHints(XY_WING, hinter);
	AHint hint = find(HINTS, "XY-Wing: G9, I8, D9 on 8");
//// Keep these, it's how I create a test case. I'm only resting for repeated.
//System.out.format("SimpleName : %s\n", hint.getClass().getSimpleName());
//System.out.format("cell       : %s\n", hint.cell);
//System.out.format("value      : %s\n", hint.value);
//System.out.format("ClueHtml   : %s\n", hint.getClueHtml(true));
//System.out.format("AquaCells  : %s\n", Frmt.csv(hint.getAquaCells(0)));
//System.out.format("Oranges    : %s\n", hint.getOranges(0));
//System.out.format("Greens     : %s\n", hint.getGreens(0));
//System.out.format("Reds       : %s\n", hint.getReds(0));
//System.out.format("RedPots    : %s\n", hint.redPots);
//System.out.format("bases      : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getBases()));
//System.out.format("covers     : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getCovers()));
	checkClue(hint, "Look for a XY-Wing on 2, 7 and <b>8</b>");
	assertEquals(grid.getCSVSet("I8, D9, G9"), hint.getAquaCells(0));
	assertEquals(newPots("D9-7, G9-27, I8-2"), hint.getOranges(0));
	assertEquals(newPots("D9-8, I8-8"), hint.getGreens(0));
	assertEquals(newPots("D8-8"), hint.getReds(0));
	assertEquals(newPots("D8-8"), hint.redPots);
	// check the Links
	{
		List<Link> expLinks = new ArrayList<>(2);
		Cell xy=grid.get("G9"), xz=grid.get("I8"), yz=grid.get("D9");
		int xValue=2, yValue=7;
		expLinks.add(new Link(xy, xValue, xz, xValue));
		expLinks.add(new Link(xy, yValue, yz, yValue));
		assertEquals(expLinks, hint.getLinks(0));
	}
	assertEquals(null, hint.getBases());
	assertEquals(null, hint.getCovers());
	assertHintEqualsStoredHintsHtml(this, hint, testName+"_1.html", REPRODUCE);
}


private static final String XYZ_WING
="7.....4...2..7..8.4.3..8..9.7.5.631..6.12..9...1..7..6...3.29...3..4.16..49..1.35" + NL
+",89,68,26,13,59,,25,13,19,,56,49,,34,56,,13,,15,,26,15,,267,27,,89,,24,,89,,,,24,35,,45,,,34,78,,78,23,89,,48,39,,25,245,,16,15,78,,56,,,47,478,58,,2578,789,,59,,,278,268,,,78,68,,278,,";
//1#C:\Users\User\Documents\NetBeansProjects\DiufSudoku\top1465.d3.mt

/** Test XYZ-Wing */
@Test
public void testXYZWing() {
	final String testName = testName(this);
	AHinter hinter = new XYWing(Tech.XYZ_Wing);
	getHints(XYZ_WING, hinter);
	AHint hint = find(HINTS, "XYZ-Wing: I7, H7, I5 on 7"); // [I8:3{278}-7]
//// Keep these, it's how I create a test case. I'm only resting for repeated.
//System.out.format("SimpleName : %s\n", hint.getClass().getSimpleName());
//System.out.format("cell       : %s\n", hint.cell);
//System.out.format("value      : %s\n", hint.value);
//System.out.format("ClueHtml   : %s\n", hint.getClueHtml(true));
//System.out.format("AquaCells  : %s\n", Frmt.csv(hint.getAquaCells(0)));
//System.out.format("Oranges    : %s\n", hint.getOranges(0));
//System.out.format("Greens     : %s\n", hint.getGreens(0));
//System.out.format("Reds       : %s\n", hint.getReds(0));
//System.out.format("RedPots    : %s\n", hint.redPots);
//System.out.format("bases      : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getBases()));
//System.out.format("covers     : %s\n", diuf.sudoku.utils.Frmt.csv(hint.getCovers()));
	checkClue(hint, "Look for a XYZ-Wing on 4, 8 and <b>7</b>");
	assertEquals(grid.getCSVSet("I7, H7, I5"), hint.getAquaCells(0));
	assertEquals(newPots("H7-4, I5-8, I7-48"), hint.getOranges(0));
	assertEquals(newPots("H7-7, I5-7"), hint.getGreens(0));
	assertEquals(newPots("I8-7"), hint.getReds(0));
	assertEquals(newPots("I8-7"), hint.redPots);
	// check the Links
	{
		List<Link> expLinks = new ArrayList<>(2);
		Cell xy=grid.get("I7"), xz=grid.get("H7"), yz=grid.get("I5");
		int xValue=4, yValue=8;
		expLinks.add(new Link(xy, xValue, xz, xValue));
		expLinks.add(new Link(xy, yValue, yz, yValue));
		assertEquals(expLinks, hint.getLinks(0));
	}
	assertEquals(null, hint.getBases());
	assertEquals(null, hint.getCovers());
	assertHintEqualsStoredHintsHtml(this, hint, testName+"_1.html", REPRODUCE);
}

}